"""
  femagtools.sm
  ~~~~~~~~~~~~~
  wound-rotor synchronous machine (EESM) electrical circuit model

  Copyright 2022: Semafor Informatik & Energie AG, Switzerland
"""
import logging
import warnings
import numpy as np
import scipy.optimize as so
import scipy.interpolate as ip
from .utils import skin_resistance, wdg_resistance, betai1, iqd, KTH
from .. import parstudy, windings
import femagtools.bch

EPS = 1e-13

eecdefaults = dict(
        zeta1=0.3,
        zeta2=0,
        gam=0.7,
        kh=2,
        tcu1=20,
        tcu2=20,
        rotor_mass=0,
        kfric_b=1,
        kpfe = 1 # iron loss factor
)

logger = logging.getLogger('sm')
logging.captureWarnings(True)


def parident(workdir, engine, machine,
             magnetizingCurves, condMat,
             **kwargs):
        """return dict of parameters of equivalent circuit for
        electrically excited synchronous machines

        arguments:
        machine -- dict() with machine parameters
        magnetizingCurves -- list of dict() with BH curves
        condMat -- list of dict() with conductor material properties

        optional arguments:
        num_cur_steps: number of current steps (default 5)
        num_beta_steps: number of current steps (default 13)
        num_exc_steps: number of excitation current (default 7)
        speed: rotor speed in 1/s (default 160/p)
        i1_max: maximum current in A rms (default approx 3*i1nom)
        """
        da1 = machine['outer_diam']
        Q1 = machine['stator']['num_slots']
        if 'statorRotor3' in machine['stator']:
            hs = machine['stator']['statorRotor3']['slot_height']
        elif 'stator1' in machine['stator']:
            hs = machine['stator']['stator1']['slot_rf1'] - machine['stator']['stator1']['tip_rh1']
        elif 'stator4' in machine['stator']:
            hs = machine['stator']['stator4']['slot_height']
        N = machine['windings']['num_wires']
        Jmax = 15  # max current density in A/mm2

        i1_max = round(0.28*np.pi*hs*(da1+hs)/Q1/N*Jmax*1e5)*10 * \
            machine['windings'].get('num_par_wdgs', 1)

        ifnom = machine['rotor']['ifnom']
        exc_logspace = True
        if exc_logspace:
            excur = np.logspace(np.log(ifnom/10), np.log(1.5*ifnom),
                                kwargs.get("num_exc_steps", 6),
                                base=np.exp(1)).tolist()
        else:
            excur = np.linspace(ifnom/10, 1.5*ifnom,
                                kwargs.get("num_exc_steps", 6))

        parvardef = {
            "decision_vars": [
                {"values": excur, "name": "load_ex_cur"}
            ]
        }

        parvar = parstudy.List(
            workdir,  condMat=condMat,
            magnetizingCurves=magnetizingCurves)

        simulation = dict(
                calculationMode=kwargs.get('calculationMode',
                                           'ld_lq_fast'),
                wind_temp=20.0,
                i1_max=kwargs.get('i1_max', i1_max),
                maxid=0,
                minid=-i1_max,
                maxiq=i1_max,
                miniq=-i1_max,
                delta_id=i1_max/kwargs.get('num_cur_steps', 5),
                delta_iq=i1_max/kwargs.get('num_cur_steps', 5),
                beta_min=-180.0,
                beta_max=0.0,
                calc_noload=0,
                num_move_steps=kwargs.get('num_move_steps', 31),
                load_ex_cur=0.5,
                num_cur_steps=kwargs.get('num_cur_steps', 5),
                num_beta_steps=kwargs.get('num_beta_steps', 13),
                num_par_wdgs=machine['windings'].get('num_par_wdgs', 1),
                period_frac=6,
                speed=50.0)

        ###self.cleanup()  # remove previously created files in workdir
        results = parvar(parvardef, machine, simulation, engine)
        if simulation['calculationMode'] == 'ld_lq_fast':
            idname = 'ldq'
        else:
            idname = 'psidq'
        for r in results['f']:
            for k in ('hf', 'ef'):
                if k in r['lossPar']:
                    r[idname]['losses'][k] = r['lossPar'][k]
        try:
            rotor_mass = sum(results['f'][-1]['weights'][-1])
        except KeyError:
            rotor_mass = 0  # need femag classic > rel-9.3.x-48-gca42bbd0
        b = results['f'][-1]

        losskeys = ('speed', 'ef', 'hf',
                    'styoke_hyst', 'stteeth_hyst', 'styoke_eddy',
                    'stteeth_eddy', 'rotor_hyst', 'rotor_eddy')

        # winding resistance
        try:
            r1 = machine['windings']['resistance']
        except KeyError:
            yd = machine['windings'].get('coil_span', Q1/machine['poles'])
            wdg = windings.Winding(
            {'Q': machine['stator']['num_slots'],
             'm': machine['windings']['num_phases'],
             'p': machine['poles']//2,
             'l': machine['windings']['num_layers'],
             'yd': yd})

            lfe = machine['lfe']
            g = machine['windings'].get('num_par_wdgs', 1)
            if 'dia_wire' in machine['windings']:
                aw = np.pi*machine['windings'].get('dia_wire', 1e-3)**2/4
            else:  # wire diameter from slot area
                aw = 0.75 * \
                        machine['windings'].get('cufilfact', 0.45)*np.pi*da1*hs/Q1/2/N
            r1 = wdg_resistance(wdg, N, g, aw, da1, hs, lfe)

        if simulation['calculationMode'] == 'ld_lq_fast':
                dqpars = dict(m=3, p=b['machine']['p'],
                        r1=r1,
                        r2=machine['rotor'].get('resistance', 1),
                        rotor_mass=rotor_mass, kfric_b=1,
                        ldq=[dict(
                                ex_current=b['machine']['ex_current'],
                                i1=b['ldq']['i1'],
                                beta=b['ldq']['beta'],
                                psid=b['ldq']['psid'],
                                psiq=b['ldq']['psiq'],
                                torque=b['ldq']['torque'],
                                ld=b['ldq']['ld'],
                                lq=b['ldq']['lq'],
                                losses={k: b['ldq']['losses'][k]
                                        for k in losskeys})
                             for b in results['f']])
        else:
                dqpars = dict(m=3, p=b['machine']['p'],
                    r1=r1,
                    r2=machine['rotor'].get('resistance', 1),
                    rotor_mass=rotor_mass, kfric_b=1,
                    psidq=[dict(
                            ex_current=b['machine']['ex_current'],
                            iq=b['psidq']['iq'],
                            id=b['psidq']['id'],
                            psid=b['psidq']['psid'],
                            psiq=b['psidq']['psiq'],
                            torque=b['psidq']['torque'],
                            losses={k: b['psidq']['losses'][k]
                                    for k in losskeys})
                           for b in results['f']])
        if 'current_angles' in results['f'][0]:
                dqpars['current_angles'] = results['f'][0]['current_angles']
        return dqpars

def _linsampl(exc, excl, a):
    """auxiliary func for linear sampling of nonlinear sequence
    arguments:
    exc: (list) nonlinear sequence of excitation current
    excl: (list) linear sequence of excitation current
    a: (array) matrix to be resampled"""
    z = []
    b, i = a.shape[1:]
    for x in range(b):
        for y in range(i):
            zl = ip.interp1d(exc, a[:, x, y], kind='linear')
            z.append(zl(excl))
    return np.array(z).T.reshape(len(excl), b, i)


def _splinterp(beta, i1, betax, i1x, a):
    """auxiliary function to increase resolution of array a
    using a cubic spline interpolation.
    arguments:
    beta: (list) n original up-i angles in rad
    i1: (list) m original current in A
    betax: (list) nx new up-i angles in rad
    i1x: (list) mx new currents
    a: (nxm array) to be interpolated"""
    f = ip.RectBivariateSpline(
        beta, i1, np.asarray(a),
        kx=3, ky=3).ev
    X, Y = np.meshgrid(betax, i1x)
    return f(X, Y).T


def _islinear(exc):
    return np.abs(np.sum(
        np.asarray(exc)**2 -
        np.linspace(exc[0], exc[-1], len(exc))**2)) < 1e-2


def gradient_respecting_bounds(bounds, fun, eps=1e-8):
    """bounds: list of tuples (lower, upper)"""
    def gradient(x):
        logger.info(x)
        fx = fun(x)
        grad = np.zeros(len(x))
        for k in range(len(x)):
            d = np.zeros(len(x))
            d[k] = eps if x[k] + eps <= bounds[k][1] else -eps
            grad[k] = (fun(x + d) - fx) / d[k]
        return grad
    return gradient


class SynchronousMachine(object):
    def __init__(self, eecpars, **kwargs):
        self.kth1 = KTH
        self.kth2 = KTH
        self.skin_resistance = [None, None]
        # here you can set user defined functions for calculating the skin-resistance,
        # according to the current frequency w. First function in list is for stator, second for rotor.
        # If None, the femagtools intern default implementation is used.
        # User defined functions need to have the following arguments:
        # - r0: (float) dc-resistance at 20°C
        # - w: (float)  current frequency in rad (2*pi*f)
        # - tcu: (float) conductor temperature in deg Celsius
        # - kth: (float) temperature coefficient (Default = 0.0039, Cu)
        for k in eecdefaults.keys():
            setattr(self, k, eecdefaults[k])

        for k in eecpars:
            if k not in ('ldq', 'psidq'):
                setattr(self, k, eecpars[k])

        for k in kwargs:
            setattr(self, k, kwargs[k])

        try:
            self.tfric = self.kfric_b*self.rotor_mass*30e-3/np.pi
        except AttributeError:
            self.tfric = 0

        self.fo = 50
        self.plexp = {'styoke_hyst': 1.0,
                      'stteeth_hyst': 1.0,
                      'styoke_eddy': 2.0,
                      'stteeth_eddy': 2.0,
                      'rotor_hyst': 1.0,
                      'rotor_eddy': 2.0}

    def _set_losspar(self, speed, ef, hf):
        self.fo = speed*self.p
        self.plexp = {'styoke_hyst': hf,
                      'stteeth_hyst': hf,
                      'styoke_eddy': ef,
                      'stteeth_eddy': ef,
                      'rotor_hyst': hf,
                      'rotor_eddy': ef}

    def pfric(self, n):
        """friction and windage losses"""
        return 2*np.pi*n*self.tfric

    def rstat(self, w):
        """stator resistance"""
        sr = self.skin_resistance[0]
        if sr is not None:
            return sr(self.r1, w, self.tcu1, kth=self.kth1)
        else:
            return skin_resistance(self.r1, w, self.tcu1, self.zeta1,
                                   self.gam, self.kh, kth=self.kth1)

    def rrot(self, w):
        """rotor resistance"""
        sr = self.skin_resistance[1]
        if sr is not None:
            return sr(self.r2, w, self.tcu2, kth=self.kth2)
        else:
            return skin_resistance(self.r2, w, self.tcu2, self.zeta2,
                                   0.0, 1, kth=self.kth2)

    def torque_iqd(self, iq, id, iex):
        "torque at q-d-current"
        psid, psiq = self.psi(iq, id, iex)
        return self.m*self.p/2*(psid*iq - psiq*id)

    def tloss_iqd(self, iq, id, iex, n):
        """return loss torque of d-q current, iron loss correction factor
        and friction windage losses"""
        if n > 1e-3:
            f1 = self.p*n
            plfe = self.kpfe * (self.iqd_plfe1(iq, id, iex, f1) + self.iqd_plfe2(iq, id, f1))
            return (plfe + self.pfric(n))/(2*np.pi*n)
        return 0

    def tmech_iqd(self, iq, id, iex, n):
        """return shaft torque of d-q current and speed"""
        return self.torque_iqd(iq, id, iex) - self.tloss_iqd(iq, id, iex, n)

    def uqd(self, w1, iq, id, iex):
        """return uq, ud of frequency w1 and d-q current"""
        psid, psiq = self.psi(iq, id, iex)
        r1 = self.rstat(w1)
        return r1*id + w1*psid, r1*iq - w1*psiq

    def plcu1(self, iqde, w1):
        r1 = self.rstat(w1)
        return 3/2*r1*(iqde[0]**2 + iqde[1]**2)

    def plcu2(self, iqde, w1=0):
        r2 = self.rrot(0)
        return r2*iqde[2]**2

    def culoss(self, iqde, w1=0):
        return self.plcu1(iqde, w1) + self.plcu2(iqde, 0)

    def iqd_plcu1(self, iq, id, f1):
        return self.plcu1((iq, id), 2*np.pi*f1)

    def iqd_plcu2(self, iq, id, iex):
        return self.plcu2((iq, id, iex))

    def iqd_plfe2(self, iq, id, f1):
        return np.zeros(np.asarray(iq).shape)

    def iqd_plmag(self, iq, id, f1):
        return np.zeros(np.asarray(iq).shape)

    def iqd_tmech(self, torque, n, disp=False, maxiter=500):
        """return currents for shaft torque with minimal losses"""
        if torque > 0:
            startvals = self.bounds[0][1], 0, sum(self.bounds[-1])
        else:
            startvals = -self.bounds[0][1], 0, sum(self.bounds[-1])

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            def sqrtculoss(iqde):
                pcu = self.culoss(iqde)
                return pcu

            res = so.minimize(
                self.culoss, startvals, method='SLSQP',  # trust-constr
                bounds=self.bounds,
                #            jac=gradient_respecting_bounds(self.bounds, self.culoss),
                constraints=[
                    {'type': 'eq',
                     'fun': lambda iqd: self.tmech_iqd(*iqd, n) - torque}])
                    #options={'disp': disp, 'maxiter': maxiter})
            if res['success']:
                return res.x

        logger.warning("%s: torque=%f %f, io=%s",
                       res['message'], torque, self.tmech_iqd(*startvals, n),
                       startvals)
        raise ValueError(res['message'])

    def iqd_torque(self, torque, disp=False, maxiter=500):
        """return currents for torque with minimal losses"""
        if torque > 0:
            startvals = self.bounds[0][1], 0, sum(self.bounds[-1])
        else:
            startvals = -self.bounds[0][1], 0, sum(self.bounds[-1])

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            def sqrtculoss(iqde):
                pcu = self.culoss(iqde)
                #logger.info("iqde %s --> pcu %f", iqde, pcu)
                return pcu
            res = so.minimize(
                self.culoss, startvals, method='SLSQP',  # trust-constr
                bounds=self.bounds,
                #            jac=gradient_respecting_bounds(self.bounds, self.culoss),
                constraints=[
                    {'type': 'eq',
                     'fun': lambda iqd: self.torque_iqd(*iqd) - torque}])
                #options={'disp': disp, 'maxiter': maxiter})
            if res['success']:
                return res.x
        logger.warning("%s: torque=%f %f, io=%s",
                       res['message'], torque, self.torque_iqd(*startvals),
                       startvals)
        raise ValueError(res['message'])

    def mtpa(self, i1max):
        """return iq, id, iex currents and maximum torque per current """
        T0 = self.torque_iqd(np.sqrt(2)*i1max, 0, self.bounds[-1][1])
        def i1tq(tq):
            return abs(i1max) - np.linalg.norm(self.iqd_torque(tq)[:2])/np.sqrt(2)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            tq = so.fsolve(i1tq, T0)[0]
        iq, id, iex = self.iqd_torque(tq)
        return iq, id, iex, tq

    def mtpa_tmech(self, i1max, n):
        """return iq, id, iex currents and maximum torque per current """
        T0 = self.torque_iqd(np.sqrt(2)*i1max, 0, self.bounds[-1][0])
        def i1tq(tq):
            return i1max - np.linalg.norm(self.iqd_tmech(tq, n)[:2])/np.sqrt(2)
        tq = so.fsolve(i1tq, T0)[0]
        iq, id, iex = self.iqd_tmech(tq, n)
        return iq, id, iex, tq

    def iqd_tmech_umax(self, torque, w1, u1max, log=0, **kwargs):
        """return currents and shaft torque at stator frequency and
         with minimal losses at max voltage"""
        iqde = self.iqd_tmech(torque, w1/2/np.pi/self.p)
        if np.linalg.norm(
            self.uqd(w1, *iqde)) <= u1max*np.sqrt(2):
            if log:
                log(iqde)
            return (*iqde, torque)
        beta, i1 = betai1(iqde[0], iqde[1])
        iex = iqde[2]

        def ubeta(b):
            return np.sqrt(2)*u1max - np.linalg.norm(
                         self.uqd(w1, *iqd(b, i1), iex))
        beta = -np.pi/4 if torque>0 else -3*np.pi/4
        io = *iqd(beta, i1), iex

        #    logger.debug("--- torque %g io %s", torque, io)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            n = w1/2/np.pi/self.p
            def sqrtculoss(iqde):
                pcu = self.culoss(iqde)
                #logger.info("iqde %s pcu %g", iqde, pcu)
                return pcu

            res = so.minimize(
                self.culoss, io, method='SLSQP',  # trust-constr
                bounds=self.bounds,
                constraints=[
                    {'type': 'eq',
                     'fun': lambda iqd: self.tmech_iqd(*iqd, n) - torque},
                    {'type': 'eq',
                     'fun': lambda iqd: np.linalg.norm(
                         self.uqd(w1, *iqd)) - u1max*np.sqrt(2)}])
            #if res['success']:
        if log:
            log(res.x)
        return *res.x, self.tmech_iqd(*res.x, n)
        #logger.warning("%s: w1=%f torque=%f, u1max=%f, io=%s",
        #               res['message'], w1, torque, u1max, io)
        #raise ValueError(res['message'])

    def iqd_torque_umax(self, torque, w1, u1max,
                        disp=False, maxiter=500, log=0, **kwargs):
        """return currents for torque with minimal losses"""
        iqde = self.iqd_torque(torque, disp, maxiter)
        if np.linalg.norm(
            self.uqd(w1, *iqde)) <= u1max*np.sqrt(2):
                if log:
                    log(iqde)
                return (*iqde, torque)
        io = iqde[0], 0, iqde[2]
        #    logger.debug("--- torque %g io %s", torque, io)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            def sqrtculoss(iqde):
                pcu = self.culoss(iqde)
                #logger.info("iqde %s pcu %g", iqde, pcu)
                return pcu

            res = so.minimize(
                self.culoss, io, method='SLSQP',  # trust-constr
                bounds=self.bounds,
                #options={'disp': disp, 'maxiter': maxiter},
                #            jac=gradient_respecting_bounds(self.bounds, self.culoss),
                constraints=[
                    {'type': 'eq',
                     'fun': lambda iqd: self.torque_iqd(*iqd) - torque},
                    {'type': 'eq',
                     'fun': lambda iqd: np.linalg.norm(
                         self.uqd(w1, *iqd)) - u1max*np.sqrt(2)}])
            if res['success']:

                if log:
                    log(res.x)
                return *res.x, self.torque_iqd(*res.x)
        logger.warning("%s: w1=%f torque=%f, u1max=%f, io=%s",
                       res['message'], w1, torque, u1max, io)
        raise ValueError(res['message'])

    def w1_imax_umax(self, i1max, u1max):
        """return frequency w1 and shaft torque at voltage u1max and current i1max

        Keyword arguments:
        u1max -- the maximum voltage (Vrms)
        i1max -- the maximum current (Arms)"""
        iq, id, iex, T = self.mtpa(i1max)
        n0 = np.sqrt(2)*u1max/np.linalg.norm(
                self.psi(iq, id, iex))/2/np.pi/self.p
        return self.w1_umax(u1max, iq, id, iex), T

    def w1_umax(self, u, iq, id, iex):
        """return frequency w1 at given voltage u and id, iq current

        Keyword arguments:
        u -- the maximum voltage (RMS)
        iq, id -- the d-q currents"""
        w10 = np.sqrt(2)*u/np.linalg.norm(self.psi(iq, id, iex))
        return so.fsolve(
            lambda w1: np.linalg.norm(self.uqd(w1, iq, id, iex))-u*np.sqrt(2),
            w10)[0]

    def characteristics(self, T, n, u1max, nsamples=50,
                        with_tmech=True, **kwargs):
        """calculate torque speed characteristics.
        return dict with list values of
        n, T, u1, i1, beta, cosphi, pmech, n_type

        Keyword arguments:
        T -- (float) the maximum torque in Nm
        n -- (float) the maximum speed in 1/s
        u1max -- (float) the maximum voltage in V rms
        nsamples -- (optional) number of speed samples
        with_tmech -- (optional) use friction and windage losses
        """
        iq, id, iex = self.iqd_torque(T)
        if with_tmech:
            i1max = betai1(iq, id)[1]
            if T < 0:
                i1max = -i1max
            w1type, Tf = self.w1_imax_umax(i1max, u1max)
        else:
            Tf = T
            w1type = self.w1_umax(u1max, iq, id, iex)
        logger.debug("w1type %f", w1type)
        wmType = w1type/self.p
        pmax = Tf*wmType

        def tload(wm):
            if abs(wm*Tf) < abs(pmax):
                return Tf
            return pmax/wm

        wmtab = []
        dw = 0
        wmMax = 3.5*wmType
        if n > 0:
            wmMax = min(wmMax, 2*np.pi*n)
        if wmType > wmMax:
            wmrange = sorted([0, wmMax])
            wmtab = np.linspace(0, wmMax, nsamples).tolist()
        else:
            wmrange = sorted([0, wmType, wmMax])
            nx = int(round(nsamples*wmType/wmMax))
            dw = (wmMax-wmType)/(nsamples-nx)
            wmtab = (np.linspace(0, wmType, nx).tolist() +
                     np.linspace(wmType+dw, wmMax, nsamples-nx).tolist())

        logger.info("Speed range T %f %s", Tf, wmrange)
        wmtab[0] = 0

        r = dict(u1=[], i1=[], id=[], iq=[], iex=[], T=[], cosphi=[], n=[],
                 beta=[], plfe1=[], plcu1=[], plcu2=[])
        for wm, tq in zip(wmtab, [tload(wx) for wx in wmtab]):
            w1 = wm*self.p
            if with_tmech:
                iq, id, iex, tqx = self.iqd_tmech_umax(
                        tq, w1, u1max)
            else:
                iq, id, iex, tqx = self.iqd_torque_umax(
                        tq, w1, u1max)
                tqx -= self.tfric
            uq, ud = self.uqd(w1, iq, id, iex)
            u1 = np.linalg.norm((uq, ud))/np.sqrt(2)
            f1 = w1/2/np.pi
            r['id'].append(id)
            r['iq'].append(iq)
            r['iex'].append(iex)
            r['u1'].append(u1)
            beta = np.arctan2(id, iq)
            if beta > 0:
                beta -= 2*np.pi
            i1 = np.linalg.norm((id, iq), axis=0)/np.sqrt(2.0)
            r['i1'].append(i1)
            r['beta'].append(beta/180*np.pi)
            gamma = np.arctan2(ud, uq)
            r['cosphi'].append(np.cos(gamma - beta))
            r['plfe1'].append(self.iqd_plfe1(iq, id, iex, f1))
            r['plcu1'].append(self.m*i1**2*self.rstat(w1))
            r['plcu2'].append(iex**2*self.rrot(0))
            r['T'].append(tqx)
            r['n'].append(wm/2/np.pi)

        r['plfe'] = r['plfe1']
        r['plcu'] = (np.array(r['plcu1']) + np.array(r['plcu2'])).tolist()
        r['plfw'] = [self.pfric(n) for n in r['n']]
        r['pmech'] = [2*np.pi*n*tq
                      for n, tq in zip(r['n'], r['T'])]
        pmech = np.array(r['pmech'])
        pltotal = (np.array(r['plfe1']) + np.array(r['plfw']) +
                   np.array(r['plcu1']) + np.array(r['plcu2']))
        r['losses'] = pltotal.tolist()

        if pmech.any():
            p1 = pmech + pltotal
            if np.abs(pmech[0]) < 1e-12:
                r['eta'] = [0]
                i = 1
            else:
                r['eta'] = []
                i = 0
            if np.all(abs(p1[i:]) > abs(pmech[i:])):
                r['eta'] += (pmech[i:]/(p1[i:])).tolist()
            else:
                r['eta'] += (p1[i:]/pmech[i:]).tolist()

        return r


class SynchronousMachinePsidq(SynchronousMachine):

    def __init__(self, eecpars, lfe=1, wdg=1, **kwargs):
        super(self.__class__, self).__init__(
                eecpars, **kwargs)
        self.iqrange = (eecpars['psidq'][0]['iq'][0],
                        eecpars['psidq'][0]['iq'][-1])
        self.idrange = (eecpars['psidq'][0]['id'][0],
                        eecpars['psidq'][0]['id'][-1])
        islinear = True
        iexc = [l['ex_current'] for l in eecpars['psidq']]
        id = [i/wdg for i in eecpars['psidq'][-1]['id']]
        idx = np.linspace(id[0], id[-1], 12)
        iq = [i/wdg for i in eecpars['psidq'][-1]['iq']]
        iqx = np.linspace(iq[0], iq[-1], 20)

        if _islinear(iexc):
            exc = iexc
            psid = wdg*lfe*np.array([
                    _splinterp(iq, id, iqx, idx, l['psid'])
                    for l in eecpars['psidq']])
            psiq = wdg*lfe*np.array([
                    _splinterp(iq, id, iqx, idx, l['psiq'])
                    for l in eecpars['psidq']])
        else:
            islinear = False
            nsamples = 10
            iexcl = np.linspace(iexc[0], iexc[-1], nsamples)
            exc = iexcl
            psid = wdg*lfe*_linsampl(iexc, iexcl, np.array(
                    [_splinterp(iq, id, iqx, idx, l['psid'])
                     for l in eecpars['psidq']]))
            psiq = wdg*lfe*_linsampl(iexc, iexcl, np.array(
                    [_splinterp(iq, id, iqx, idx, l['psiq'])
                     for l in eecpars['psidq']]))

        self.psidf = ip.RegularGridInterpolator(
                (exc, iqx, idx), psid,
                method='cubic', bounds_error=False, fill_value=None)
        self.psiqf = ip.RegularGridInterpolator(
                (exc, iqx, idx), psiq,
                method='cubic', bounds_error=False, fill_value=None)
        self.bounds = [(min(iq), max(iq)),
                       (min(id), 0),
                       (iexc[0], iexc[-1])]
        try:
            idname = 'psidq'
            keys = self.plexp.keys()
            if islinear:
                pfe = {k: np.array([l['losses'][k]
                                    for l in eecpars[idname]])
                       for k in keys}
            else:
                pfe = {k: _linsampl(iexc, iexcl,
                                    np.array([l['losses'][k]
                                             for l in eecpars[idname]]))
                       for k in keys}
            self._losses = {k: ip.RegularGridInterpolator(
                    (exc, iq, id), lfe*np.array(pfe[k]),
                    method='cubic', bounds_error=False, fill_value=None)
                            for k in keys}
            self._set_losspar(eecpars[idname][0]['losses']['speed'],
                          eecpars[idname][0]['losses']['ef'],
                          eecpars[idname][0]['losses']['hf'])
        except KeyError:
            logger.warning("loss map missing")
            self._losses = {k: lambda x: 0 for k in (
                'styoke_hyst', 'stteeth_hyst',
                'styoke_eddy', 'stteeth_eddy',
                'rotor_hyst', 'rotor_eddy')}


    def psi(self, iq, id, iex):
        """return psid, psiq of currents iq, id"""
        try:
            return self.psidf((iex, iq, id)), self.psiqf((iex, iq, id))
        except ValueError as ex:
            logger.error(iex, iq, id)
            raise ex

    def plfe1(self, iq, id, iex, f1):
        return np.sum([
            self._losses[k]((iex, iq, id))*(f1/self.fo)**self.plexp[k][0]
            for k in ('styoke_eddy', 'styoke_hyst',
                      'stteeth_eddy', 'stteeth_hyst')], axis=0)

    def iqd_plfe1(self, iq, id, iex, f1):
        return self.plfe1(iq, id, iex, f1)


class SynchronousMachineLdq(SynchronousMachine):
    def __init__(self, eecpars, lfe=1, wdg=1, **kwargs):
        super(self.__class__, self).__init__(eecpars, **kwargs)
        self.betarange = (eecpars['ldq'][0]['beta'][0]/180*np.pi,
                          eecpars['ldq'][0]['beta'][-1]/180*np.pi)
        self.i1range = (0, eecpars['ldq'][0]['i1'][-1])
        islinear = True
        iexc = [l['ex_current'] for l in eecpars['ldq']]
        i1 = [i/wdg for i in eecpars['ldq'][-1]['i1']]
        i1x = np.linspace(i1[0], i1[-1], 12)
        beta = [b*np.pi/180 for b in eecpars['ldq'][-1]['beta']]
        betax = np.linspace(beta[0], beta[-1], 20)

        if _islinear(iexc):
            exc = iexc
            psid = wdg*lfe*np.array([
                    _splinterp(beta, i1, betax, i1x, l['psid'])
                    for l in eecpars['ldq']])
            psiq = wdg*lfe*np.array([
                    _splinterp(beta, i1, betax, i1x, l['psiq'])
                    for l in eecpars['ldq']])
        else:
            islinear = False
            nsamples = 10
            iexcl = np.linspace(iexc[0], iexc[-1], nsamples)
            exc = iexcl
            psid = wdg*lfe*_linsampl(iexc, iexcl, np.array(
                    [_splinterp(beta, i1, betax, i1x, l['psid'])
                     for l in eecpars['ldq']]))
            psiq = wdg*lfe*_linsampl(iexc, iexcl, np.array(
                    [_splinterp(beta, i1, betax, i1x, l['psiq'])
                     for l in eecpars['ldq']]))

        # extrapolate outside range
        self.psidf = ip.RegularGridInterpolator(
                (exc, betax, i1x), np.sqrt(2)*psid,
                method='cubic',
                bounds_error=False, fill_value=None)
        self.psiqf = ip.RegularGridInterpolator(
                (exc, betax, i1x), np.sqrt(2)*psiq,
                method='cubic'
                , bounds_error=False, fill_value=None)
        i1max = np.sqrt(2)*(max(i1))
        self.bounds = [(np.cos(min(beta))*i1max, i1max),
                       (-i1max, 0),
                       (iexc[0], iexc[-1])]
        keys = self.plexp.keys()
        try:
            idname = 'ldq'
            if islinear:
                pfe = {k: np.array([l['losses'][k]
                                    for l in eecpars[idname]])
                       for k in keys}
            else:
                pfe = {k: _linsampl(iexc, iexcl,
                                    np.array([l['losses'][k]
                                             for l in eecpars[idname]]))
                       for k in keys}
        # fill value with nan outside range
            self._losses = {k: ip.RegularGridInterpolator(
                    (exc, beta, i1), lfe*np.array(pfe[k]),
                    method='cubic', bounds_error=False, fill_value=None)
                            for k in keys}
            self._set_losspar(eecpars[idname][0]['losses']['speed'],
                              eecpars[idname][0]['losses']['ef'],
                              eecpars[idname][0]['losses']['hf'])
        except ValueError:
            logger.warning("loss map missing")
            self._losses = {k: lambda x: 0 for k in (
                'styoke_hyst', 'stteeth_hyst',
                'styoke_eddy', 'stteeth_eddy',
                'rotor_hyst', 'rotor_eddy')}

    def psi(self, iq, id, iex):
        """return psid, psiq of currents iq, id"""
        beta = np.arctan2(id, iq)
        if beta > 0:
            beta -= 2*np.pi
        i1 = np.linalg.norm((id, iq), axis=0)/np.sqrt(2.0)
        try:
            return self.psidf((iex, beta, i1)), self.psiqf((iex, beta, i1))
        except ValueError as ex:
            logger.error("iex %s iq %f id %f beta %f i1 %f",
                         iex, iq, id, beta, i1)
            raise ex

    def plfe1(self, beta, i1, iex, f1):
        return np.sum([
            self._losses[k]((iex, beta, i1))*(f1/self.fo)**self.plexp[k][0]
            for k in ('styoke_eddy', 'styoke_hyst',
                      'stteeth_eddy', 'stteeth_hyst')], axis=0)
    def iqd_plfe1(self, iq, id, iex, f1):
        beta = np.arctan2(id, iq)
        if np.isscalar(beta):
            if beta > 0:
                beta -= 2*np.pi
        else:
            beta[beta > 0] -= 2*np.pi
        i1 = np.linalg.norm((id, iq), axis=0)/np.sqrt(2.0)
        return self.plfe1(beta, i1, iex, f1)


if __name__ == '__main__':
    import sys
    import json
    import femagtools.plot
    import matplotlib.pyplot as plt
    with open(sys.argv[1]) as fp:
        eecpar = json.load(fp)
    if 'ldq' in eecpar:
        m = SynchronousMachineLdq(eecpar)
    else:
        m = SynchronousMachinePsidq(eecpar)
    T = 240
    u1max = 163
    nmax = 1000
    r = m.characteristics(T, 0, u1max)
    femagtools.plot.characteristics(r, 'SynchronousMachine')
    plt.show()



#!/usr/bin/env python
#
import unittest
import sys
sys.path.insert(0, '../..')
from moo.population import Population
from moo.problem import Problem
from moo.algorithm import Nsga2

import numpy as np
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d as mpl
import matplotlib.cm as cm


class FesProblem(Problem):
    def __init__(self):
        super(FesProblem, self).__init__(4, 0, 2)
        
    def objfun(self, x):
        D = len(x)
        return (sum([abs(x[i]-np.exp(((i+1.)/D) ** 2)/3) ** 0.5
                    for i in range(D)]),
                sum([(x[i]-0.5*np.cos(10*(i+1)*np.pi/D)-0.5) ** 2
                    for i in range(D)]))


def surface():
    "show a nice 3d plot just for demo purposes"
    prob = FesProblem()
    steps = (15, 15)
    lower = (0.0, 0.0)
    upper = (1.0, 1.0)
    domain = [np.linspace(l, u, s)
              for s, l, u in zip(steps, lower, upper)]

    X = domain[0]
    Y = domain[1]

    L = [len(d) for d in domain]
    s = []
    e = 1
    for d in domain:
        L = L[1:]
        LS = np.prod(L)
        s.append(np.repeat(d.tolist()*LS, e))
        if len(L) > 0:
            e = e*L[0]

    z = np.array([prob.objfun(x) for x in np.array(s).T]).T[1]
    z = np.reshape(z, steps)

    fig = plt.figure()
    ax = mpl.Axes3D(fig)
    surf = ax.plot_surface(X, Y, z,
                           rstride=1,
                           cstride=1,
                           cmap=cm.jet,
                           linewidth=0)
    plt.show()


class AlgorithmTest(unittest.TestCase):
    def test_evolve(self):
        prob = FesProblem()
        pop = Population(prob, 20)
        pop.eval()

        F = np.array([ind.cur_f for ind in pop.individuals]).T
        plt.scatter(F[0], F[1], c="b", s=50, alpha=0.5)
        F = np.array([ind.cur_f
                      for ind in pop.individuals if ind.rank == 0]).T
        plt.scatter(F[0], F[1], c="g", s=100, alpha=0.5)
    
        algo = Nsga2()
        for i in range(25):
            newpop = algo.evolve(pop)
            newpop.eval()
            pop.merge(newpop)
            
        F = np.array([ind.cur_f for ind in pop.individuals]).T
        plt.scatter(F[0], F[1], c="r", s=200, alpha=0.5)
    
        plt.xlabel("$f^{(1)}$")
        plt.ylabel("$f^{(2)}$")
        plt.savefig('moo.png')

        self.assertEqual(pop.compute_norm_dist(), 1.0)
        self.assertEqual(pop.dom_count, [0]*pop.size())
        self.assertEqual(pop.pareto_rank, [0]*pop.size())

if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
#
import unittest
import femagtools.fsl
import femagtools.magnet
import copy
import re

modelpars = dict(
    name="PM 130 L4",
    outer_diam=0.13,
    bore_diam=0.07,
    lfe=0.1,
    poles=4,
    stator=dict(
        num_slots=12,
        mcvkey_yoke="3",
        num_slots_gen=3,
        nodedist=1.5,
        rlength=1.0),
    windings=dict(
        num_phases=3,
        num_layers=1,
        num_wires=4,
        coil_span=3))

feapars = dict(
    lfe=0.1,
    speed=50.0,
    current=10.0,
    nu_move_steps=49,
    num_cur_steps=5,
    angl_i_up=0,
    optim_i_up=0,
    wind_temp=60.0,
    magn_temp=60.0,
    eval_force=0,
    calc_fe_loss=1,
    cog_move_steps=90,
    num_layers=1,
    slot_indul=0,
    skew_angle=0.0,
    culength=1.4,
    num_par_wdgs=1,
    cufilfact=0.45,
    num_skew_steps=0)


class FslBuilderTest(unittest.TestCase):

    def setUp(self):
        self.m = copy.deepcopy(modelpars)
        self.builder = femagtools.fsl.Builder()

    def tearDown(self):
        self.m = None
        self.builder = None

    def test_stator1(self):
        self.m['stator']['stator1'] = dict(
            tooth_width=0.009,
            slot_rf1=0.002,
            tip_rh1=0.002,
            tip_rh2=0.002,
            slot_width=0.003)
        model = femagtools.MachineModel(self.m)
        fsl = self.builder.create_stator_model(model)
        self.assertEqual(len(fsl), 27)

    def test_stator2(self):
        self.m['stator']['stator2'] = dict(
            slot_width=0.009,
            slot_t1=0.002,
            slot_t2=0.002,
            slot_t3=0.002,
            corner_width=0.002,
            slot_depth=0.003)
        model = femagtools.MachineModel(self.m)
        fsl = self.builder.create_stator_model(model)
        self.assertEqual(len(fsl), 23)

    def test_stator3(self):
        self.m['stator']['statorRotor3'] = dict(
            slot_h1=0.002,
            slot_h2=0.004,
            middle_line=0,
            tooth_width=0.009,
            wedge_width2=0.0,
            wedge_width1=0.0,
            slot_top_sh=0,
            slot_r2=0.002,
            slot_height=0.02,
            slot_r1=0.003,
            slot_width=0.003)
        model = femagtools.MachineModel(self.m)
        fsl = self.builder.create_stator_model(model)
        self.assertEqual(len(fsl), 41)

    def test_stator4(self):
        self.m['stator']['stator4'] = dict(
            slot_height=0.1,
            slot_h1=1e-3,
            slot_h2=0,
            slot_h3=2e-3,
            slot_h4=3e-4,
            slot_r1=11e-3,
            slot_width=22e-3,
            wedge_width1=111e-5,
            wedge_width2=222e-5,
            wedge_width3=333e-5)
        model = femagtools.MachineModel(self.m)
        fsl = self.builder.create_stator_model(model)
        self.assertEqual(len(fsl), 39)

    def test_statorBG(self):
        self.m['stator']['statorBG'] = dict(
            yoke_diam_ins=0.0344,
            slottooth=0.0,
            tip_rad=0.0,
            middle_line=1,
            slot_h1=1e-3,
            slot_r1=0,
            slot_h3=2e-3,
            slot_r2=3e-4,
            tooth_width=3.2e-3,
            slot_width=22e-3)

        model = femagtools.MachineModel(self.m)
        fsl = self.builder.create_stator_model(model)
        self.assertEqual(len(fsl), 33)

    def test_magnetSector(self):
        self.m['magnet'] = dict(
            magnetSector=dict(
                magn_height=0.005,
                magn_width_pct=0.8,
                condshaft_r=0.0591,
                magn_rfe=0.0,
                magn_len=1.0,
                magn_shape=0.0,
                bridge_height=0.0,
                bridge_width=0.0,
                magn_ori=2,
                magn_type=1,
                magn_num=1))
        model = femagtools.MachineModel(self.m)
        fsl = self.builder.create_magnet_model(model)
        self.assertEqual(len(fsl), 30)

    def test_magnetIron(self):
        self.m['magnet'] = dict(
            magnetIron=dict(
                magn_height=0.005,
                magn_width=0.008,
                gap_ma_iron=0,
                air_triangle=5,
                iron_height=0.001,
                magn_rem=1.2,
                condshaft_r=0.0591,
                magn_ori=1,
                bridge_height=0,
                bridge_width=0,
                iron_shape=0))
        model = femagtools.MachineModel(self.m)
        fsl = self.builder.create_magnet_model(model)
        # print('\n'.join(fsl))
        self.assertEqual(len(fsl), 26)

    def test_magnetIron2(self):
        self.m['magnet'] = dict(
            magnetIron2=dict(
                magn_height=0.005,
                magn_width=0.008,
                gap_ma_iron=0,
                air_triangle=1,
                iron_height=0.001,
                magn_rem=1.2,
                condshaft_r=0.006,
                magn_ori=1,
                gap_ma_right=1e-3,
                gap_ma_left=2e-3,
                iron_shape=0))
        model = femagtools.MachineModel(self.m)
        fsl = self.builder.create_magnet_model(model)
        self.assertEqual(len(fsl), 28)

    def test_magnetIron3(self):
        self.m['magnet'] = dict(
            magnetIron3=dict(
                magn_height=0.005,
                magn_width=0.008,
                gap_ma_iron=0,
                iron_bfe=3e-3,
                magn_num=1,
                air_triangle=1,
                iron_height=0.001,
                condshaft_r=0.006,
                magn_ori=1,
                gap_ma_right=1e-3,
                gap_ma_left=2e-3,
                iron_shape=0))
        model = femagtools.MachineModel(self.m)
        fsl = self.builder.create_magnet_model(model)
        self.assertEqual(len(fsl), 27)

    def test_magnetIron4(self):
        self.m['magnet'] = dict(
            magnetIron4=dict(
                magn_height=0.005,
                magn_width=0.008,
                gap_ma_iron=0,
                iron_bfe=3e-3,
                magn_num=1,
                air_space_h=1e-3,
                iron_height=0.001,
                corner_r=0.006,
                magn_ori=1,
                magn_di_ra=1e-3,
                air_sp_ori=1,
                iron_shape=0))
        model = femagtools.MachineModel(self.m)
        fsl = self.builder.create_magnet_model(model)
        self.assertEqual(len(fsl), 27)

    def test_magnetIron5(self):
        self.m['magnet'] = dict(
            magnetIron5=dict(
                magn_height=0.005,
                magn_width=0.008,
                gap_ma_iron=0,
                iron_bfe=3e-3,
                magn_num=1,
                air_space_h=1e-3,
                iron_height=0.001,
                corner_r=0.006,
                air_space_b=1e-3,
                magn_di_ra=1e-3,
                air_sp_ori=1,
                iron_shape=0))
        model = femagtools.MachineModel(self.m)
        fsl = self.builder.create_magnet_model(model)
        self.assertEqual(len(fsl), 27)

    def test_magnetIronV(self):
        self.m['magnet'] = dict(
            magnetIronV=dict(
                magn_height=0.005,
                magn_width=0.008,
                gap_ma_iron=0,
                iron_hs=3e-3,
                magn_num=1,
                magn_rem=1.2,
                air_triangle=1,
                iron_height=0.001,
                condshaft_r=0.006,
                magn_angle=130,
                iron_shape=0))
        model = femagtools.MachineModel(self.m)
        fsl = self.builder.create_magnet_model(model)
        self.assertEqual(len(fsl), 28)

    def test_magnetFC2(self):
        self.m['magnet'] = dict(
            magnetFC2=dict(
                magn_height=0.005,
                magn_width=0.008,
                yoke_height=5e-3,
                iron_h1=3e-3,
                iron_h2=2e-3,
                iron_hp=2e-3,
                iron_b=2e-3,
                magn_num=1,
                iron_bfe=0.001,
                iron_bfo=0.001,
                iron_shape=0))
        model = femagtools.MachineModel(self.m)
        fsl = self.builder.create_magnet_model(model)
        self.assertEqual(len(fsl), 25)

    def test_rot_hsm(self):
        self.m['rotor'] = dict(
            rot_hsm=dict(
                gap_pol_shaft=1e-3,
                core_height=0.02,
                pole_height=0.016,
                pole_rad=0.042,
                core_width2=0.04,
                core_width1=0.04,
                pole_width_r=0.05,
                pole_width=0.052,
                slot_width=0.002,
                slot_height=0.002,
                damper_diam=0.004,
                damper_div=0.007
            ))
        model = femagtools.MachineModel(self.m)
        fsl = self.builder.create_rotor_model(model, condMat=[])
        self.assertEqual(len(fsl), 33)

    def test_fe_losses(self):
        model = femagtools.MachineModel(self.m)
        fsl = self.builder.create_fe_losses(model)
        self.assertFalse(fsl)
        self.m['stator']['ffactor'] = 1.1
        model = femagtools.MachineModel(self.m)
        fsl = self.builder.create_fe_losses(model)
        self.assertEqual(len(fsl), 21)
        ffact = [float(f.split('=')[1])
                 for f in fsl if f.startswith('m.ffactor')][0]
        self.assertEqual(ffact, self.m['stator']['ffactor'])
        feloss = [f.split('"')[1]
                  for f in fsl if f.find('pre_models') > 0][0]
        self.assertEqual(feloss, 'FE-Losses-1')

    def test_run_models(self):
        feapars['calculationMode'] = "cogg_calc"
        fsl = self.builder.create_analysis(feapars)
        self.assertEqual(len(fsl), 27)

        feapars['calculationMode'] = "pm_sym_fast"
        fsl = self.builder.create_analysis(feapars)
        self.assertEqual(len(fsl), 34)

        feapars['calculationMode'] = "mult_cal_fast"
        fsl = self.builder.create_analysis(feapars)
        self.assertEqual(len(fsl), 32)

        feapars['calculationMode'] = "torq_calc"
        fsl = self.builder.create_analysis(feapars)
        self.assertEqual(len(fsl), 25)

    def test_run_existing_model(self):
        model = femagtools.MachineModel('data/magnsec')
        feapars['calculationMode'] = "cogg_calc"
        fsl = self.builder.create(model, feapars)
        self.assertEqual(len(fsl), 67)

    def test_create_plots(self):
        pars = copy.deepcopy(feapars)
        pars['calculationMode'] = "pm_sym_fast"
        pars['plots'] = ['field_lines', 'Babs']
        fsl = self.builder.create_analysis(pars)
        field_lines = re.findall(r'field_lines\(([^)]*)\)', ''.join(fsl))
        self.assertEqual(len(field_lines), 1)
        self.assertEqual(int(field_lines[0].split(',')[-1]), 20)

        colorgrad = re.findall(r'color_gradation\(([^)]*)\)', ''.join(fsl))
        self.assertEqual(len(field_lines), 1)
        min, max = [int(l) for l in colorgrad[0].split(',')[4:6]]
        self.assertEqual(min, 0)
        self.assertEqual(max, 0)

        pars['plots'] = [('field_lines', 10), ('Babs', 0.0, 2.0)]
        fsl = self.builder.create_analysis(pars)
        field_lines = re.findall(r'field_lines\(([^)]*)\)', ''.join(fsl))
        self.assertEqual(len(field_lines), 1)
        self.assertEqual(int(field_lines[0].split(',')[-1]), 10)

        colorgrad = re.findall(r'color_gradation\(([^)]*)\)', ''.join(fsl))
        self.assertEqual(len(field_lines), 1)
        min, max = [float(l) for l in colorgrad[0].split(',')[4:6]]
        self.assertEqual(min, 0.0)
        self.assertEqual(max, 2.0)

    def test_readfsl(self):
        content = [
            'dshaft = 360 --shaft diameter',
            'hm  = 38 -- magnet height',
            'bm = 122 -- magnet width',
            'ws = 10  -- slot width',
            'lfe = 224',
            '-- calculate slot height, angle and pole pairs',
            'hs = (da2-dy2)/2 - bm   ',
            'alpha = math.pi/p/2     -- slot angle',
            'p   = m.num_poles/2',
            'x = {}',
            'y = {}',
            '-- Berechnung der Koordinaten',
            'x[1],y[1] = pr2c(dy2/2, 0)',
            'x[2],y[2] = pr2c(da2/2, 0)',
            'x[3],y[3] = pr2c(da2/2, alpha - math.atan2(ws,(da2/2)))',
            'x[4],y[4] = pr2c(da2/2-hs, alpha - math.atan2(ws,(da2/2 - hs)))',
            'nc_line(x[1], y[1], x[2], y[2], 0)']
        result = self.builder.read(content)
        self.assertEqual(len(result['parameter']), 4)
        for p in result['parameter']:
            self.assertTrue(p['key'] in ['dshaft', 'hm', 'bm', 'ws'])

    def test_readfslexp(self):
        from io import StringIO
        with StringIO("""
        varname1 = 1. -- equals 1.0
        varname2 = .1 -- equals 0.1
        varname3 = 1e3 -- equals 1000""") as f:
            fsltmpl = self.builder.read(f)
        self.assertEqual(fsltmpl['parameter'], [
            {'key': 'varname1', 'value': '1.', 'comment': 'equals 1.0'},
            {'key': 'varname2', 'value': '.1', 'comment': 'equals 0.1'},
            {'key': 'varname3', 'value': '1e3', 'comment': 'equals 1000'}])

    def test_gen_winding(self):
        model = femagtools.MachineModel(self.m)

        fsl = self.builder.create_gen_winding(model)
        self.assertEqual(len(fsl), 21)

        model.windings['leak_dist_wind'] = dict(
            perimrad=1,
            vbendrad=1,
            endheight=1,
            wiredia=1)
        fsl = self.builder.create_gen_winding(model)
        self.assertEqual(len(fsl), 34)

        model.windings.pop('leak_dist_wind')
        model.windings['leak_evol_wind'] = dict(
            evol1rad=1,
            evol2rad=1,
            botlevel=1,
            toplevel=1,
            endheight=1,
            evolbend=1,
            wiredia=1)
        fsl = self.builder.create_gen_winding(model)
        self.assertEqual(len(fsl), 39)

        model.windings.pop('leak_evol_wind')
        model.windings['leak_tooth_wind'] = dict(
            endheight=1,
            bendrad=1,
            wiredia=1)
        fsl = self.builder.create_gen_winding(model)
        self.assertEqual(len(fsl), 35)

    def test_create_model_with_magnet_material(self):
        magnetmat = [dict(
            name='M45',
            rlen=0.9,
            remanenc=1.1,
            relperm=1.04,
            spmaweight=7.4,
            temcoefbr=-0.0015,
            temcoefhc=-0.0013,
            magncond=625000.0
        )]

        machine = dict(
            name="PM 886 32",
            lfe=0.224,
            poles=32,
            outer_diam=0.886,
            bore_diam=0.76,
            inner_diam=0.4956,
            airgap=0.007,
            external_rotor=1,

            stator=dict(
                num_slots=120,
                rlength=1.0,
                stator4=dict(
                    slot_height=0.035,
                    slot_h1=0.002,
                    slot_h2=0.0,
                    slot_h3=0.004,
                    slot_h4=0.0,
                    slot_width=0.01,
                    slot_r1=0.0,
                    wedge_width1=0.01,
                    wedge_width2=0.01,
                    wedge_width3=0.01)
            ),

            magnet=dict(
                material='M45',
                magnetSector=dict(
                    magn_num=1,
                    magn_height=0.014,
                    magn_width_pct=0.85,
                    condshaft_r=0.0,
                    magn_rfe=0.0,
                    magn_len=1.0,
                    magn_shape=0.0,
                    bridge_height=0.0,
                    bridge_width=0.0,
                    magn_ori=1,
                    magn_type=1
                )
            ),

            windings=dict(
                material='Cu',
                num_phases=3,
                num_wires=5,
                coil_span=1,
                num_layers=2)
        )
        model = femagtools.MachineModel(machine)
        magnets = femagtools.magnet.Magnet(magnetmat)
        condMat = femagtools.magnet.Magnet([dict(name='Cu', elconduct=56e6)])
        fsl = self.builder.create_model(model, magnets, condMat)
        self.assertEqual(len(fsl), 186)
        brem = [l.strip() for l in fsl
                if l.split('=')[0].strip() == 'm.remanenc']
        self.assertEqual(brem[-1].split('=')[-1].strip(),
                         str(magnetmat[0]['remanenc']))
        rlen = [l.strip() for l in fsl
                if l.split('=')[0].strip() == 'm.rlen']
        self.assertEqual(rlen[0].split('=')[-1].strip(),
                         str(100*magnetmat[0]['rlen']))


if __name__ == '__main__':
    unittest.main()



import pytest
from femagtools import nc
from femagtools import isa7


@pytest.fixture
def model():
    filename = 'src/tests/data/minimal.nc'
    return nc.read(filename)

@pytest.fixture
def pm():
    filename = 'src/tests/data/pm_data.nc'
    return nc.read(filename)


def test_read(model):
    assert len(model.points) == 39
    assert len(model.lines) == 39
    assert model.points[0].x == pytest.approx(0.008339,
                                            abs=1e-5)


def test_objects(model):
    assert len(model.nodes) == 1729
    assert len(model.nodechains) == 98
    assert len(model.elements) == 822
    assert len(model.superelements) == 3
    assert len(model.get_subregion('asdf').elements()) == 756

def test_no_such_subregion(model):
    with pytest.raises(ValueError) as excinfo:
        n = 'foo'
        model.get_subregion(n)
    assert str(excinfo.value) == (
        'no such subregion "{}" in this model'.format(n))

def test_points(model):
    p = model.points[0].x, model.points[0].y
    p_xy = model.points[0].xy

    assert p == p_xy == pytest.approx((0.008339, 0.011702), abs=1e-5)


def test_lines(model):
    ln = model.lines[0]

    assert type(ln.p1) == isa7.Point
    assert type(ln.p2) == isa7.Point
    assert ln.p1.xy == pytest.approx((0.008339, 0.011702), abs=1e-5)
    assert ln.p2.xy == pytest.approx((0.007388, 0.007190), abs=1e-5)


def test_nodes(model):
    nd = model.nodes[-1]

    assert type(nd) == isa7.Node
    assert nd.xy == (nd.x, nd.y) == pytest.approx(
        (0.008787, 0.007736), abs=1e-5)


def test_nodechains(model):
    nc = model.nodechains[0]

    assert type(nc) == isa7.NodeChain
    assert type(nc.nodes[0]) == isa7.Node


def test_nodechains_reverse(model):
    for ndc in model.nodechains:
        assert ndc.reverse().nodes == ndc.nodes[::-1]


def test_elements(model):
    el = model.elements[0]

    assert type(el) == isa7.Element
    for v in el.vertices:
        assert type(v) == isa7.Node
    assert el.vertices[0].xy == pytest.approx((0.002202, 0.003197),
                                              abs=1e-5)

def test_airgap_center_elements(pm):
    assert len(pm.airgap_center_elements) == 180

def test_areas(pm):
    a = pm.get_areas()
    assert a[0]['iron'] == pytest.approx(0.003550, abs=1e-5)
    assert a[0]['slots'] ==  pytest.approx(0.00184, abs=1e-5)
    assert a[0]['magnets'] == 0.0
    assert a[1]['iron'] == pytest.approx(0.0007222, abs=1e-5)
    assert a[1]['slots'] == 0
    assert a[1]['magnets'] == 0

def test_calc_iron_loss(pm):
    import numpy as np
    def pfe(Bxnu, Bynu, fnu, losscoeffs):
        basfrq = 50
        basind = 1.5
        b21 = np.linalg.norm((Bxnu, Bynu), axis=0)
        b2 = (b21/basind)
        f = fnu/basfrq
        hch = f
        hcw = f**2
        ph = f*b2
        pw = f**2*b2**2
        pexc = f**1.5*b2**1.5
        return ph, pw, pexc

    icur = 0
    ibeta = 0
    pfe = pm.calc_iron_loss(icur, ibeta, pfe)
    assert pfe['Rüc'] == [0, 0, 0]
    assert pfe['Stat'] == pytest.approx([46.95,  5.61, 15.92], abs=1e-2)


def test_superelements(model):
    se = model.superelements[0]

    assert type(se) == isa7.SuperElement
    for el in se.elements:
        assert type(el) == isa7.Element
    for ndc in se.nodechains:
        assert type(ndc) == isa7.NodeChain
        assert ndc.key == se.nc_keys[se.nodechains.index(ndc)]


@pytest.fixture
def disp_stat():
    filename = 'src/tests/data/test_disp_stat.nc'
    return nc.read(filename)


def test_subregions(disp_stat):
    sr = disp_stat.subregions[0]

    assert type(sr) == isa7.SubRegion
    assert sr.name == "Stat"
    for se in sr.superelements:
        assert type(se) == isa7.SuperElement
    for ndc in sr.nodechains:
        assert type(ndc) == isa7.NodeChain


def test_windings(disp_stat):
    wd = disp_stat.windings[0]

    assert type(wd) == isa7.Winding
    assert wd.name == "Stra"
    for sr in wd.subregions:
        assert type(sr) == isa7.SubRegion
    assert wd.num_turns == 100


def test_magnet_super_elements(disp_stat):
    sekeys = [se.key for se in disp_stat.magnet_super_elements()]
    assert sekeys == [98, 101, 74, 77, 80, 83, 86, 89, 92, 95]



import numpy as np
import matplotlib.pyplot as plt
import logging


logger = logging.getLogger("femagtools.plot.nc")


def spel(isa, with_axis=False, ax=0):
    """plot super elements of I7/ISA7 model
    Args:
      isa: Isa7 object
    """
    from matplotlib.patches import Polygon
    if ax == 0:
        ax = plt.gca()
    ax.set_aspect('equal')
    for se in isa.superelements:
        ax.add_patch(Polygon([n.xy
                              for nc in se.nodechains
                              for n in nc.nodes],
                             color=isa.color[se.color], lw=0))

    ax.autoscale(enable=True)
    if not with_axis:
        ax.axis('off')


def mesh(isa, with_axis=False, ax=0):
    """plot mesh of I7/ISA7 model
    Args:
      isa: Isa7 object
    """
    from matplotlib.lines import Line2D
    if ax == 0:
        ax = plt.gca()
    ax.set_aspect('equal')
    for el in isa.elements:
        z = np.array([v.xy for v in el.vertices])
        pts = np.vstack((z, z[0])).T
        ax.add_line(Line2D(pts[0], pts[1],
                           color='b', ls='-', lw=0.25))

    # for nc in isa.nodechains:
    #    pts = [list(i) for i in zip(*[(n.x, n.y) for n in nc.nodes])]
    #    ax.add_line(Line2D(pts[0], pts[1], color="b", ls="-", lw=0.25,
    #                       marker=".", ms="2", mec="None"))

    # for nc in isa.nodechains:
    #    if nc.nodemid is not None:
    #        plt.plot(*nc.nodemid.xy, "rx")

    ax.autoscale(enable=True)
    if not with_axis:
        ax.axis('off')


def _contour(ax, title, elements, values, label='', isa=None):
    from matplotlib.patches import Polygon
    from matplotlib.collections import PatchCollection
    if ax == 0:
        ax = plt.gca()
    ax.set_aspect('equal')
    ax.set_title(title)
    if isa:
        for se in isa.superelements:
            ax.add_patch(Polygon([n.xy
                                  for nc in se.nodechains
                                  for n in nc.nodes],
                                 color='gray', alpha=0.1, lw=0))
    valid_values = np.logical_not(np.isnan(values))
    patches = np.array([Polygon([v.xy for v in e.vertices])
                       for e in elements])[valid_values]
    # , cmap=matplotlib.cm.jet, alpha=0.4)
    p = PatchCollection(patches, alpha=1.0, match_original=False)
    p.set_array(np.asarray(values)[valid_values])
    ax.add_collection(p)
    cb = plt.colorbar(p, shrink=0.9)
    for patch in np.array([Polygon([v.xy for v in e.vertices],
                                   fc='white', alpha=1.0)
                           for e in elements])[np.isnan(values)]:
        ax.add_patch(patch)
    if label:
        cb.set_label(label=label)
    ax.autoscale(enable=True)
    ax.axis('off')


def demag(isa, ax=0):
    """plot demag of NC/I7/ISA7 model
    Args:
      isa: Isa7/NC object
    """
    emag = [e for e in isa.elements if e.is_magnet()]
    demag = np.array([e.demagnetization(isa.MAGN_TEMPERATURE) for e in emag])
    _contour(ax, f'Demagnetization at {isa.MAGN_TEMPERATURE} °C (max -{np.max(demag):.1f} kA/m)',
             emag, demag, '-H / kA/m', isa)
    logger.info("Max demagnetization %f", np.max(demag))


def demag_pos(isa, pos=-1, icur=-1, ibeta=-1, ax=0):
    """plot demag of NC/I7/ISA7 model at rotor position
    Args:
      isa: Isa7/NC object
      pos: rotor position in degree (maximum h if -1)
      icur: cur amplitude index or last index if -1
      ibeta: beta angle index or last index if -1
    """
    emag = [e for e in isa.elements if e.is_magnet()]
    demag = np.array([isa.demagnetization(e, icur, ibeta)[1]
                      for e in emag])
    if pos>=0:
        for i, x in enumerate(isa.pos_el_fe_induction):
            if x >= pos/180*np.pi:
                break
    else:
        demagmax = np.max(demag, axis=0)
        i = np.argmax(demagmax)
        x = isa.pos_el_fe_induction[i]

    hpol = demag[:, i]
    hpol[hpol == 0] = np.nan
    _contour(ax, f'Demagnetization at pos. {round(x/np.pi*180):.1f}°,'
    f'{isa.MAGN_TEMPERATURE} °C (max -{np.max(hpol):.1f} kA/m)',
             emag, hpol, '-H / kA/m', isa)
    logger.info("Max demagnetization %f kA/m", np.nanmax(hpol))


def __elements_of_subreg(isa, subreg):
    if subreg:
        if isinstance(subreg, list):
            sr = subreg
        else:
            sr = [subreg]
        for s in sr:
            for e in isa.get_subregion(s).elements():
                yield e
    else:
        for e in isa.elements:
            yield e


def flux_density(isa, subreg=[], ax=0):
    """plot flux density of NC/I7/ISA7 model
    Args:
      isa: Isa7/NC object
    """
    elements = [e for e in __elements_of_subreg(isa, subreg)]
    fluxd = np.array([np.linalg.norm(e.flux_density()) for e in elements])
    _contour(ax, f'Flux Density T (max {np.max(fluxd):.1f} T)',
             elements, fluxd)
    logger.info("Max flux dens %f", np.max(fluxd))


def max_flux_density(isa, subreg=[], icur=-1, ibeta=-1, ax=0):
    """plot max flux density of each element of NC/I7/ISA7 model
    Args:
      isa: Isa7/NC object
    """
    elements = [e for e in __elements_of_subreg(isa, subreg)]
    bmax = []
    for e in elements:
        fd = isa.flux_density(e, icur, ibeta)
        bmax.append(np.max(
            np.linalg.norm((fd['bx'], fd['by']), axis=0)))
    fluxd = np.array(bmax)
    _contour(ax, f'Max Flux Density T (max {np.max(fluxd):.1f} T)',
             elements, fluxd)
    logger.info("Max flux dens %f", np.max(fluxd))


def min_flux_density(isa, subreg=[], icur=-1, ibeta=-1, ax=0):
    """plot min flux density of each element of NC/I7/ISA7 model
    Args:
      isa: Isa7/NC object
    """
    elements = [e for e in __elements_of_subreg(isa, subreg)]
    bmin = []
    for e in elements:
        fd = isa.flux_density(e, icur, ibeta)
        bmin.append(np.min(
            np.linalg.norm((fd['bx'], fd['by']), axis=0)))

    fluxd = np.array(bmin)
    _contour(ax, f'Min Flux Density T (max {np.max(fluxd):.1f} T)',
             elements, fluxd)
    logger.info("Max flux dens %f (element %d)",
                np.max(fluxd), elements[np.argmax(fluxd)].key)


def flux_density_pos(isa, ipos, subreg=[], icur=-1, ibeta=-1, ax=0):
    """plot flux density at rotor pos for each element of NC/I7/ISA7 model
    Args:
      isa: Isa7/NC object
    """
    elements = [e for e in __elements_of_subreg(isa, subreg)]
    b = []
    for e in elements:
        fd = isa.flux_density(e, icur, ibeta)
        b.append(fd['bx'][ipos])
    fluxd = np.array(b)
    pos = isa.pos_el_fe_induction[ipos]*180/np.pi
    isa.rotate(isa.pos_el_fe_induction[ipos])
    _contour(ax, f'Flux Density T at {pos:.1f}° (max {np.max(fluxd):.1f} T)',
             elements, fluxd)
    logger.info("Max flux dens %f", np.max(fluxd))
    isa.rotate(0)


def airgap_flux_density_pos(isa, ipos, icur=-1, ibeta=-1, ax=0):
    """plot flux density at rotor pos for each element of NC/I7/ISA7 model
    Args:
      isa: Isa7/NC object
    """
    bx = []
    by = []
    pos = []
    for e in isa.airgap_center_elements:
        fd = isa.flux_density(e, icur, ibeta)
        bx.append(fd['bx'][ipos])
        by.append(fd['by'][ipos])
        pos.append(np.arttan2(e.center[1], e.center[0])/np.pi*100)
    if ax == 0:
        ax = plt.gca()
    ax.plot(pos, bx)
    ax.plot(pos, by)
    ax.grid()

    logger.info("Max flux dens %f", np.max(np.abs(bx)))


def loss_density(isa, subreg=[], ax=0):
    """plot loss density of NC/I7/ISA7 model
    Args:
      isa: Isa7/NC object
    """
    elements = [e for e in __elements_of_subreg(isa, subreg)]
    lossd = np.array([e.loss_density*1e-3 for e in elements])
    _contour(ax, 'Loss Density kW/m³', elements, lossd)



import pytest
from femagtools import ts


@pytest.fixture
def losses():
    dirname = 'src/tests/data/zzz_pm_model_ts_results_1'
    modelname = 'src/tests/data/zzz_pm_model_ts'
    return ts.Losses(modelname, dirname)


def test_read(losses):
    e = losses.ohm_lossenergy(0.0, 0.0)
    assert len(e) == 6
    assert {'key': 1, 'name': 'stfe', 'losses': 0.0} == e[0]
    assert {'key': 2, 'name': 'rofe', 'losses': 0.0} == e[1]
    assert {'key': 3, 'name': 'wefe', 'losses': 0.0} == e[2]
    assert pytest.approx(
        {'key': 4, 'name': '', 'losses': 0.001913}, abs=1e-5) == e[3]
    assert pytest.approx(
        {'key': 5, 'name': '', 'losses': 0.001967}, abs=1e-5) == e[4]
    assert pytest.approx(
        {'key': 6, 'name': '', 'losses': 0.001886}, abs=1e-5) == e[5]
    p = losses.ohm_powerlosses()
    assert len(p) == 6
    assert {'key': 1, 'name': 'stfe', 'losses': 0.0} == p[0]
    assert {'key': 2, 'name': 'rofe', 'losses': 0.0} == p[1]
    assert {'key': 3, 'name': 'wefe', 'losses': 0.0} == p[2]
    assert pytest.approx(
        {'key': 4, 'name': '', 'losses': 0.09809}, abs=1e-5) == p[3]
    assert pytest.approx(
        {'key': 5, 'name': '', 'losses': 0.10102}, abs=1e-5) == p[4]
    assert pytest.approx(
        {'key': 6, 'name': '', 'losses': 0.09673}, abs=1e-5) == p[5]
    p = losses.ohm_powerlosses_fft()
    assert len(p) == 6
    assert {'key': 1, 'name': 'stfe', 'losses': 0.0} == p[0]
    assert {'key': 2, 'name': 'rofe', 'losses': 0.0} == p[1]
    assert {'key': 3, 'name': 'wefe', 'losses': 0.0} == p[2]
    assert pytest.approx(
        {'key': 4, 'name': '', 'losses': 0.101086}, abs=1e-5) == p[3]
    assert pytest.approx(
        {'key': 5, 'name': '', 'losses': 0.101086}, abs=1e-5) == p[4]
    assert pytest.approx(
        {'key': 6, 'name': '', 'losses': 0.101086}, abs=1e-5) == p[5]



"""
    femagtools.engine.multiproc
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Creating and managing multicore/multiprocessing jobs



    :authors: R. Tanner, N. Mauchle
"""
import platform
import multiprocessing
import subprocess
import time
import os
import re
import threading
import pathlib
import logging
from .job import Job
import femagtools.config as cfg
try:
    from subprocess import DEVNULL
except ImportError:
    DEVNULL = open(os.devnull, 'wb')

logger = logging.getLogger(__name__)

numpat = re.compile(r'([+-]?\d+(?:\.\d+)?(?:[eE][+-]\d+)?)\s*')

class LicenseError(Exception):
    pass

class ProtFile:
    def __init__(self, dirname, num_cur_steps):
        self.size = 0
        self.looplen = 0
        self.cur_steps = [1, num_cur_steps]
        self.n = 0
        self.num_loops = 0
        self.dirname = dirname
        self.name = 'samples'

    def percent(self):
        if self.looplen > 0:
            return min(100 * self.n / self.looplen, 100)
        return 0

    def update(self):
        p = list(pathlib.Path(self.dirname).glob('*.PROT'))
        if p:
            buf = ''
            if self.size < p[0].stat().st_size:
                with p[0].open() as fp:
                    fp.seek(self.size)
                    buf = fp.read()
            return self.append(buf)
        return ''

    def append(self, buf):
        self.size += len(buf)
        for line in [l.strip() for l in buf.split('\n') if l]:
            if line.startswith('Loop'):
                self.n = 0
                try:
                    cur_steps = self.cur_steps[self.num_loops]
                except IndexError:
                    cur_steps = 1
                x0, x1, dx, nbeta = [float(f)
                                     for f in re.findall(numpat, line)][:4]
                move_steps = round((x1-x0)/dx+1)
                beta_steps = int(nbeta)
                self.looplen = cur_steps*beta_steps*move_steps
                self.num_loops += 1
            elif (line.startswith('Cur') or
                  line.startswith('Id')):
                self.n += 1
            elif line.startswith('Number movesteps Fe-Losses'):
                return ''
            elif line.startswith('begin'):
                self.name = line.split()[1].strip()

        return f'{self.percent():3.1f}%'  # {self.n}/{self.looplen}'


class ProgressLogger(threading.Thread):
    def __init__(self, dirs, num_cur_steps, timestep):
        threading.Thread.__init__(self)
        self.dirs = dirs
        self.num_cur_steps = num_cur_steps
        self.running = False
        self.timestep = timestep

    def run(self):
        self.running = True
        protfiles = [ProtFile(d, self.num_cur_steps)
                     for d in self.dirs]
        while self.running:
            time.sleep(self.timestep)
            logmsg = [p.update() for p in protfiles]
            summary = [l  # f'<{i}> {l}'
                       for i, l in enumerate(logmsg)
                       if l]
            if summary:
                labels = set([p.name for p in protfiles])
                logger.info('%s: %s',
                            ', '.join(labels),
                            ', '.join(summary))
            else:
                logger.info('collecting FE losses ...')
                return

    def stop(self):
        self.running = False


def run_femag(cmd, workdir, fslfile):
    """Start the femag command as subprocess.

    :internal:

    Args:
        cmd: (list) The program (executable image) to be run
        workdir: The workdir where the calculation files are stored
        fslfile: The name of the start file (usually femag.fsl)
    """
    with open(os.path.join(workdir, "femag.out"), "wb") as out, \
            open(os.path.join(workdir, "femag.err"), "wb") as err:
        try:
            proc = subprocess.Popen(cmd + ['-b', fslfile],
                                    shell=False,
                                    stdin=DEVNULL,
                                    stdout=out,
                                    stderr=err,
                                    cwd=workdir)
            logger.info('%s (pid %d, workdir %s)', cmd, proc.pid, workdir)
            # write pid file
            with open(os.path.join(workdir, 'femag.pid'), 'w') as pidfile:
                pidfile.write("{}\n".format(proc.pid))

            # wait
            proc.wait()
            os.remove(os.path.join(workdir, 'femag.pid'))

            #logger.info("Finished pid: %d return %d",
            #            proc.pid, proc.returncode)
            #return proc.returncode
        except OSError as e:
            logger.error("Starting process failed: %s, Command: %s", e, cmd)
            raise

    # raise License Error
    if proc.returncode != 0:
        with open(os.path.join(workdir, "femag.err"), "r") as err:
            for line in err:
                if 'license' in line:
                    raise LicenseError(line)

    logger.info("Finished pid: %d return %d", proc.pid, proc.returncode)
    return proc.returncode


class Engine:
    """The MultiProc engine uses a pool of local calculation processes.

    This is more or less a decorator for the `Python multiprocessing Module
    <https://docs.python.org/3.6/library/multiprocessing.html>`_

    Args:
        cmd: the program (executable image) to be run
            (femag dc is used if None)
        process_count: number of processes (cpu_count() if None)
        progress_timestep: time step in seconds for progress log messages if > 0)
    """

    def __init__(self, **kwargs):
        self.process_count = kwargs.get('process_count', None)
        cmd = kwargs.get('cmd', '')
        if cmd:
            self.cmd = [cmd]
            if platform.system() == 'Windows':
                self.cmd.append('-m')

        self.progressLogger = 0
        self.progress_timestep = kwargs.get('timestep', 5)

    def create_job(self, workdir):
        """Create a FEMAG :py:class:`Job`

        Args:
            workdir: The workdir where the calculation files are stored

        Return:
            FEMAG :py:class:`Job`
        """
        self.job = Job(workdir)
        return self.job

    def submit(self, extra_result_files=[]):
        """Starts the FEMAG calculation(s) with the internal
        :py:meth:`multiproc.run_femag` function

        Return:
            length of started tasks
        """
        # must check if cmd is set:
        args = []
        if platform.system() == 'Windows':
            args.append('-m')

        try:
            for t in self.job.tasks:
                t.cmd = self.cmd
        except AttributeError:
            for t in self.job.tasks:
                t.cmd = [cfg.get_executable(
                    t.stateofproblem)] + args

        self.pool = multiprocessing.Pool(self.process_count)
        self.tasks = [self.pool.apply_async(run_femag,
                                            args=(t.cmd,
                                                  t.directory,
                                                  t.fsl_file))
                      for t in self.job.tasks]
        self.pool.close()

        if (self.progress_timestep and
                self.job.num_cur_steps):
            self.progressLogger = ProgressLogger(
                [t.directory for t in self.job.tasks],
                num_cur_steps=self.job.num_cur_steps,
                timestep=self.progress_timestep)
            self.progressLogger.start()
        return len(self.tasks)

    def join(self):
        """Wait until all calculations are finished

        Return:
            list of all calculations status (C = Ok, X = error)
        """
        exitcodes = [task.get() for task in self.tasks]
        status = []
        for t, ec in zip(self.job.tasks, exitcodes):
            t.status = 'C'
            if ec != 0:
                t.status = 'X'
                errmsg = pathlib.Path(t.directory) / 'femag.err'
                if errmsg.exists():
                    t.errmsg = errmsg.read_text()
                    logger.error(t.errmsg)
            status.append(t.status)
        if self.progressLogger:
            self.progressLogger.stop()
        return status

    def terminate(self):
        logger.info("terminate Engine")
        if self.progressLogger:
            self.progressLogger.stop()
        # terminate pool
        try:
            self.pool.terminate()
            self.pool.close()
        except AttributeError:
            logger.warn("%s", e)



# -*- coding: utf-8 -*-
"""
    femagtools.losscoeffs
    ~~~~~~~~~~~~~~~~~~~~~

    Fitting methods for loss coeffs



"""
import numpy as np
import scipy.optimize as so


def pfe_jordan(f, B, ch, fh, cw, fw, fb, fo, Bo):
    return (ch*(f/fo)**fh + cw*(f/fo)**fw)*(B/Bo)**fb


def pfe_steinmetz(f, B, cw, fw, fb, fo, Bo):
    return cw*(f/fo)**fw * (B/Bo)**fb


def fitsteinmetz(f, B, losses, Bo, fo, alpha0=1.0):
    """fit coeffs of
    losses(f,B)=cw*(f/fo)**alfa*(B/Bo)**beta
    returns (cw, alfa, beta)
    """
    if np.isscalar(f):
        fbx = [[f]*len(losses), B]
        y = losses
        fitp, cov = so.curve_fit(
            lambda x, cw, beta: pfe_steinmetz(
                x[0], x[1], cw, alpha0, beta, fo, Bo),
            fbx, y, (1.0, 2.0))
        fitp = np.insert(fitp, 1, alpha0)

    else:
        pfe = losses
        z = []
        for i, fx in enumerate(f):
            if fx:
                if isinstance(B[0], float):
                    z += [(fx, bx, y)
                          for bx, y in zip(B, pfe[i])
                          if isinstance(y, float)]
                else:
                    z += [(fx, bx, y)
                          for bx, y in zip(B[i], pfe[i]) if y]

        fbx = np.array(z).T[0:2]
        y = np.array(z).T[2]

        fitp, cov = so.curve_fit(
            lambda x, cw, alpha, beta: pfe_steinmetz(
                x[0], x[1], cw, alpha, beta, fo, Bo),
            fbx, y, (1.0, 1.0, 2.0))
    return fitp


def fitjordan(f, B, losses, Bo, fo):
    """fit coeffs of
    losses(f,B)=(ch*(f/fo)**alpha + ch*(f/fo)**beta)*(B/Bo)**gamma
    returns (ch, alpha, cw, beta, gamma)
    """
    pfe = losses
    z = []
    for i, fx in enumerate(f):
        if fx:
            if isinstance(B[0], float):
                z += [(fx, bx, y)
                      for bx, y in zip(B, pfe[i])
                      if y]
            else:
                z += [(fx, bx, y)
                      for bx, y in zip(B[i], pfe[i])
                      if y]

    fbx = np.array(z).T[0:2]
    y = np.array(z).T[2]
    fitp, cov = so.curve_fit(lambda x, ch, alpha, cw, beta, gamma: pfe_jordan(
        x[0], x[1], ch, alpha, cw, beta, gamma, fo, Bo),
        fbx, y, (1.0, 1.0, 1.0, 2.0, 1.0))
    return fitp




"""
  Copyright (c) 2011 Semafor Informatik & Energie AG

  This file is part of opal.

  opal is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  This software is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with this software; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""



"""
    femagtools.plot.char
    ~~~~~~~~~~~~~~~~~~~~

    Creating characteristics plots


"""
import numpy as np
import matplotlib.pyplot as plt
import logging


def mtpa(pmrel, i1max, title='', projection='', ax=0):
    """create a line or surface plot with torque and mtpa curve"""
    nsamples = 10
    i1 = np.linspace(0, i1max, nsamples)
    iopt = np.array([pmrel.mtpa(x) for x in i1]).T

    iqmax, idmax = pmrel.iqdmax(i1max)
    iqmin, idmin = pmrel.iqdmin(i1max)

    if projection == '3d':
        nsamples = 50
    else:
        if iqmin == 0:
            iqmin = 0.1*iqmax
    id = np.linspace(idmin, idmax, nsamples)
    iq = np.linspace(iqmin, iqmax, nsamples)

    torque_iqd = np.array(
        [[pmrel.torque_iqd(x, y)
          for y in id] for x in iq])
    if projection == '3d':
        from .bch import idq_torque
        ax = idq_torque(id, iq, torque_iqd, ax)
        ax.plot(iopt[1], iopt[0], iopt[2],
                color='red', linewidth=2, label='MTPA: {0:5.0f} Nm'.format(
                    np.max(iopt[2][-1])))
    else:
        if ax == 0:
            ax = plt.gca()

        ax.set_aspect('equal')
        x, y = np.meshgrid(id, iq)
        CS = ax.contour(x, y, torque_iqd, 6, colors='k')
        ax.clabel(CS, fmt='%d', inline=1)

        ax.set_xlabel('Id/A')
        ax.set_ylabel('Iq/A')
        ax.plot(iopt[1], iopt[0],
                color='red', linewidth=2, label='MTPA: {0:5.0f} Nm'.format(
                    np.max(iopt[2][-1])))
        ax.grid()

    if title:
        ax.set_title(title)
    ax.legend()


def mtpv(pmrel, u1max, i1max, title='', projection='', ax=0):
    """create a line or surface plot with voltage and mtpv curve"""
    w1 = pmrel.w2_imax_umax(i1max, u1max)
    nsamples = 20
    if projection == '3d':
        nsamples = 50

    iqmax, idmax = pmrel.iqdmax(i1max)
    iqmin, idmin = pmrel.iqdmin(i1max)
    id = np.linspace(idmin, idmax, nsamples)
    iq = np.linspace(iqmin, iqmax, nsamples)
    u1_iqd = np.array(
        [[np.linalg.norm(pmrel.uqd(w1, iqx, idx))/np.sqrt(2)
          for idx in id] for iqx in iq])
    u1 = np.mean(u1_iqd)
    imtpv = np.array([pmrel.mtpv(wx, u1, i1max)
                      for wx in np.linspace(w1, 20*w1, nsamples)]).T

    if projection == '3d':
        from .bch import idq_torque
        torque_iqd = np.array(
            [[pmrel.torque_iqd(x, y)
              for y in id] for x in iq])
        ax = idq_torque(id, iq, torque_iqd, ax)
        ax.plot(imtpv[1], imtpv[0], imtpv[2],
                color='red', linewidth=2)
    else:
        if ax == 0:
            ax = plt.gca()
        ax.set_aspect('equal')
        x, y = np.meshgrid(id, iq)
        CS = ax.contour(x, y, u1_iqd, 4, colors='b')  # linestyles='dashed')
        ax.clabel(CS, fmt='%d', inline=1)

        ax.plot(imtpv[1], imtpv[0],
                color='red', linewidth=2,
                label='MTPV: {0:5.0f} Nm'.format(np.max(imtpv[2])))
        # beta = np.arctan2(imtpv[1][0], imtpv[0][0])
        # b = np.linspace(beta, 0)
        # ax.plot(np.sqrt(2)*i1max*np.sin(b), np.sqrt(2)*i1max*np.cos(b), 'r-')

        ax.grid()
        ax.legend()
    ax.set_xlabel('Id/A')
    ax.set_ylabel('Iq/A')
    if title:
        ax.set_title(title)


def characteristics(char, title=''):
    fig, axs = plt.subplots(2, 2, figsize=(10, 8),
                            layout='tight',
                            sharex=True)
    if title:
        fig.suptitle(title)

    n = np.array(char['n'])*60
    punit = 'kW'
    k = 1e-3
    if max(char['pmech']) > 1e6:
        punit = 'MW'
        k = 1e-6
    pmech = np.array(char['pmech'])*k
    tunit = 'Nm'
    if max(char['T']) > 1e3:
        tunit = 'kNm'
        tq = np.array(char['T'])*1e-3
    else:
        tq = np.array(char['T'])

    axs[0, 0].plot(n, tq, 'C0-', label='Torque')
    axs[0, 0].set_ylabel(f"Torque / {tunit}")
    axs[0, 0].grid()
    axs[0, 0].legend(loc='center left')
    ax1 = axs[0, 0].twinx()
    ax1.plot(n, pmech, 'C1-', label='P mech')
    ax1.set_ylabel(f"Power / {punit}")
    ax1.legend(loc='lower center')

    axs[0, 1].plot(n[1:], np.array(char['u1'][1:]), 'C0-', label='Voltage')
    axs[0, 1].set_ylabel("Voltage / V",)
    axs[0, 1].grid()
    axs[0, 1].legend(loc='center left')
    ax2 = axs[0, 1].twinx()
    ax2.plot(n[1:], char['cosphi'][1:], 'C1-', label='Cos Phi')
    ax2.set_ylabel("Cos Phi")
    ax2.legend(loc='lower right')

    if 'id' in char:
        axs[1, 0].plot(n, np.array(char['id']), label='Id')
    if 'iq' in char:
        axs[1, 0].plot(n, np.array(char['iq']), label='Iq')
    axs[1, 0].plot(n, np.array(char['i1']), label='I1')
    axs[1, 0].set_xlabel("Speed / rpm")
    axs[1, 0].set_ylabel("Current / A")
    axs[1, 0].legend(loc='center left')
    if 'beta' in char:
        ax3 = axs[1, 0].twinx()
        ax3.plot(n, char['beta'], 'C3-', label='Beta')
        ax3.set_ylabel("Beta / °")
        ax3.legend(loc='center right')
    axs[1, 0].grid()
    try:
        plfe = np.array(char['plfe'])*1e-3
    except KeyError:
        plfe = np.array(char['plfe1'])*1e-3
    try:
        plcu = np.array(char['plcu'])*1e-3
    except KeyError:
        plcu = np.array(char['plcu1'])*1e-3
    pl = np.array(char['losses'])*1e-3
    axs[1, 1].plot(n, plcu, 'C0-', label='Cu Losses')
    axs[1, 1].plot(n, plfe, 'C1-', label='Fe Losses')
    try:
        if char['plfw'] and char['plfw'][-1] > 0:
            plfw = np.array(char['plfw'])*1e-3
            axs[1, 1].plot(n, plfw, 'C2-', label='Friction + Windage')
    except KeyError:
            pass
    axs[1, 1].set_ylabel("Losses / kW")
    axs[1, 1].legend(loc='center left')
    axs[1, 1].grid()
    axs[1, 1].set_xlabel("Speed / rpm")
    ax4 = axs[1, 1].twinx()
    ax4.plot(n[1:-1], char['eta'][1:-1], 'C3-', label="Eta")
    ax4.legend(loc='upper center')
    ax4.set_ylabel("Efficiency")

    return fig


def get_nT_boundary(n, T):
    """extract boundary from n,T lists"""
    n0 = n[0]
    t0 = T[0]
    #      braking, driving
    bnd = [[(n0, t0)], []]
    for nx, tx in zip(n, T):
        if tx < t0:
            bnd[1].append((n0, t0))
            bnd[0].append((nx, tx))
            n0 = nx
        t0 = tx
    bnd[1].append((nx, tx))
    return np.array(bnd[0] + bnd[1][::-1])

def normalize10(v, **kwargs):
    """
    Normalizes the input-array using the nearest (ceiling) power of 10.

    Arguments:
        v: array to normalize

    Returns:
        normalized array
        normalisation factor (power of 10)
    """
    n_type = kwargs.get('n_type', 'abs') # 'norm')
    r_type = kwargs.get('r_type', 'round') # 'floor')
    if n_type == 'norm':
        norm = np.log10(np.linalg.norm(v))
    elif n_type == 'abs':
        norm = np.log10(np.max(np.abs(v)))
    else:
        raise AttributeError('Unknown norm-type. Allowed values are: "norm", "abs".')

    if r_type == 'floor':
        norm = 10 ** np.floor(norm)
    elif r_type == 'ceil':
        norm = 10 ** np.ceil(norm)
    elif r_type == 'round':
        norm = 10 ** np.round(norm)
    else:
        raise AttributeError('Unknown rounding type. Allowed values are: "floor", "ceil", "round".')
    # norm = 10**(np.ceil(np.log10(np.max(np.abs(v)))))
    # norm = 10 ** (np.ceil(np.log10(np.linalg.norm(v))))
    # norm = 10 ** (np.floor(np.log10(np.max(np.abs(v)))))
    # norm = 10 ** (np.floor(np.log10(np.linalg.norm(v))))
    return v / norm, norm


def plot_contour(speed, torque, z, ax, title='', levels=[],
                 clabel=True, cmap='YlOrRd', cbar=False, **kwargs):
    """ contour plot of speed, torque, z values
    Arguments:
    levels: (list of floats)
    clabel: contour labels if True
    cmap: colour map
    cbar: (bool) create color bar if True (default False)

    Note: the x and y axes are scaled

    returns tricontourf, xscale, yscale
    """
    from matplotlib.path import Path
    from matplotlib.patches import PathPatch
    x = 60*np.asarray(speed)
    y = np.asarray(torque)

    x, xscale = normalize10(x, **kwargs)
    y, yscale = normalize10(y, **kwargs)

    if not levels:
        if max(z) <= 1:
            if max(z) > 0.96:
                levels = [0.5, 0.75, 0.8, 0.84,
                          0.89, 0.92, 0.94, 0.96, 0.97]
            else:
                levels = [0.25, 0.5, 0.75, 0.8, 0.84,
                          0.88, 0.9, 0.92, 0.94, 0.96]

            if max(z) > levels[-1]:
                levels.append(np.ceil(max(z)*100)/100)
        else:
            levels = 14
    #
    ax.spines['top'].set_color('none')
    ax.spines['right'].set_color('none')

    cont = ax.tricontour(x, y, z,
                         linewidths=0.4, levels=levels,
                         colors='k')
    contf = ax.tricontourf(x, y, z,
                           levels=levels, cmap=cmap)

    clippath = Path(get_nT_boundary(x, y))
    patch = PathPatch(clippath, facecolor='none')
    ax.add_patch(patch)
    for c in cont.collections:
        c.set_clip_path(patch)
    #ax.plot(x, y, "k.", ms=3)
    if clabel:
        ax.clabel(cont, inline=True, colors='k', fontsize=8, inline_spacing=0)

    for c in contf.collections:
        c.set_clip_path(patch)

    if xscale > 1:
        def format_fn(tick_val, tick_pos):
            return round(xscale*tick_val)
        ax.xaxis.set_major_formatter(format_fn)
    if yscale > 1:
        def format_fn(tick_val, tick_pos):
            return round(yscale*tick_val)
        ax.yaxis.set_major_formatter(format_fn)

    ax.set_ylabel('Torque / Nm')
    ax.set_xlabel('Speed / rpm')
    ax.set_title(title)
    if cbar:
        cfig = ax.get_figure()
        cfig.colorbar(contf, ax=ax,
                      orientation='vertical')
    return contf, xscale, yscale


def efficiency_map(rmap, ax=0, title='Efficiency Map', clabel=True,
                   cmap='YlOrRd', levels=None, cbar=False):
    if ax == 0:
        fig, ax = plt.subplots(figsize=(12, 12))
    return plot_contour(rmap['n'], rmap['T'], rmap['eta'], ax,
                        title=title, clabel=clabel, cmap=cmap,
                        levels=levels, cbar=cbar)


def losses_map(rmap, ax=0, title='Losses Map / kW', clabel=True,
               cmap='YlOrRd', cbar=False):
    if ax == 0:
        fig, ax = plt.subplots(figsize=(12, 12))
    return plot_contour(rmap['n'], rmap['T'], np.asarray(rmap['losses'])/1e3, ax,
                        title=title, levels=14, clabel=clabel,
                        cmap=cmap, cbar=cbar)



# -*- coding: utf-8 -*-
"""
    femagtools.airgap
    ~~~~~~~~~~~~~~~~~

    Read airgap dat file


"""
import numpy as np
import logging
from . import utils

logger = logging.getLogger(__name__)


def fft(pos, b, pmod=0):
    """calculate fft spectrum of flux density and return samples,
    values, amplitude and phase of base harmonic

    Arguments:
      pos: (list of floats) sample positions
      b: (list of floats) flux density values
      pmod: number of poles in model (ignored if 0)
    """
    r = utils.fft(pos, b, pmod)
    Bamp = r['a']
    alfa0 = r['alfa0']
    T0 = r['T0']
    npoles = 2*int(np.ceil(360/T0))
    logger.info("flux density: %s poles B amp %f ",
                npoles, r['a'])
    return dict(Bamp=Bamp, npoles=npoles,
                phi0=alfa0,
                pos=pos.tolist(),
                B=b.tolist(),
                nue=np.arange(0, 9*npoles).tolist(),
                B_nue=r['nue'],
                B_fft=(Bamp*np.cos(2*np.pi*pos/T0+alfa0)).tolist(),
                Bi=r['yi'],
                phi=np.linspace(pos[0], 360+pos[0], len(r['yi'])).tolist())


def read(filename, pmod=0):
    """read dat file with columns (phi, Br, Bphi)
    returns samples, values, amplitude and phase of base harmonic

    Args:
      filename: the name of the file to be processed
      pmod: number of poles in model (ignored if 0)
    """
    bag = np.loadtxt(filename).T
    if len(bag) < 3:
        logger.warn("%s has incomplete content", filename)
        return(dict())

    return fft(bag[0], bag[1], pmod)



# -*- coding: utf-8 -*-
"""
    femagtools.hxy
    ~~~~~~~~~~~~~~

    Reading HXY files (EXPERIMENTAL)

"""

import numpy as np
from collections import defaultdict

# K-means clustering
class point():
    def __init__(self, index, k, coord):
        self.index = index
        self.coord = coord
        self.k = k

def make_k_mapping(points):
    region = defaultdict(list)
    for p in points:
        region[p.k] = region[p.k] + [p.coord]
    return region

def calc_k_means(region):
    return [np.mean(region[k], axis=0) for k in region]

def update_k(points, means):
    for p in points:
        dists = [np.linalg.norm(m - p.coord) for m in means]
        p.k = np.argmin(dists)

def fit(points, epochs=10):
    for e in range(epochs):
        region = make_k_mapping(points)
        means = calc_k_means(region)
        update_k(points, means)
    return means, points

def evaluate(points):
    region = make_k_mapping(points)
    means = calc_k_means(region)
    dists = [np.linalg.norm(means[p.k]-p.coord) for p in points]
    return np.mean(dists)

def llf_(y, X, pr):
    # return maximized log likelihood
    nobs = float(X.shape[0])
    nobs2 = nobs / 2.0
    nobs = float(nobs)
    resid = y - pr
    ssr = np.sum((resid)**2)
    llf = -nobs2*np.log(2*np.pi) - nobs2*np.log(ssr / nobs) - nobs2
    return llf


def aic(y, X, pr, p):
    # return aic metric
    llf = llf_(y, X, pr)
    return -2*llf+2*p


def readSections(f):
    section = []
    movepos = False
    for line in f:
        if line.startswith(' MOVE POSITION'):
            movepos = True
            if section:
                # skip empty lines
                i = 0
                try:
                    while not section[i]:
                        i = i+1
                except IndexError:
                    i = i-1
                yield section[i:]
                section = []
        if movepos:
            section.append(line.strip())
    yield section


def read(filename, num_magnets):
    """read hxy file and return values grouped to magnets"""
    hxy = []
    with open(filename, encoding='latin1', errors='ignore') as f:
        for s in readSections(f):
            pos = float(s[0].split()[-1])
            num = np.array([[float(x) for x in l.split()] for l in s[5:] if l])
            hxy.append({'pos': pos, 'e': num[:, :2], 'hxy': num[:, 2:4],
                        'bxy': num[:, 4:6], 'mxy':num[:, 6:]})
        K = num_magnets
        points = [point(i, np.random.randint(0,K), xy)
                  for i, xy in enumerate(hxy[0]['e'])]
        new_means, new_points = fit(points)
        # move values to magnets:
        magnets = [{'e': [p.coord for p in new_points if p.k == k],
                    'pos': [], 'hxy': [], 'bxy': [], 'mxy': []}
                   for k in range(K)]
        hkeys = ['hxy', 'bxy', 'mxy']
        for i, h in enumerate(hxy):  # all positions
            for mag in magnets:
                mag['pos'].append(h['pos'])
                m = [{k: [] for k in hkeys}
                     for kk in range(K)]
            for p in new_points:  # all elements
                for k in hkeys:
                    m[p.k][k].append(h[k][p.k])
            for mk, magk in zip(m, magnets):
                for k in hkeys:
                    magk[k].append(mk[k])
        for mag in magnets:
            for k in ['e'] + hkeys:
                mag[k] = np.array(mag[k])
            mag['havg'] = []
            mag['hmax'] = []
            for hpos in mag['hxy']:
                h = np.abs(np.linalg.norm(hpos, axis=1))
                mag['havg'].append(np.mean(h))
                mag['hmax'].append(np.max(h))

        # Note dimension of hkeys is (positions x elements x 2)

    return magnets



# -*- coding: utf-8 -*-
"""
    dxfsl
    ~~~~~

    renders geometry


"""
import numpy as np
import logging

logger = logging.getLogger(__name__)


#############################
#       DumpRenderer        #
#############################

class DumpRenderer(object):
    def __init__(self, name):
        self.model = name

    def circle(self, center, radius):
        self.content.append(u'Crcle({}, {}, {})'.format(center[0],
                                                        center[1],
                                                        radius))

    def arc(self, startangle, endangle, center, radius):
        p1 = (center[0] + radius*np.cos(startangle),
              center[1] + radius*np.sin(startangle))
        p2 = (center[0] + radius*np.cos(endangle),
              center[1] + radius*np.sin(endangle))
        self.content.append(
            u"Arc({}, {}, {}, {}, {}, {})".format(
                p1[0], p1[1], p2[0], p2[1],
                center[0], center[1]))

    def line(self, p1, p2):
        self.content.append(
            u"Line({}, {}, {}, {})".format(
                p1[0], p1[1], p2[0], p2[1]))

    def render(self, geom):
        '''create file with elements'''
        incl_bnd = True
        self.content = []
        handled = set()

        # collect all areas
        for i, area in enumerate(geom.areas(incl_bnd)):
            self.content.append('-- {}'.format(i))
            for e in area:
                if e not in handled:
                    e.render(self)
                    handled.add(e)
        self.content.append('--')
        geom.remove_areas(incl_bnd)
        # and all remaining edges and circles
        for circle in geom.circles():
            circle.render(self)

        for e1, e2, attr in geom.g.edges(data=True):
            attr['object'].render(self)

        return self.content



"""
    femagtools.plot.airgap
    ~~~~~~~~~~~~~~~~~~~~~~

    Creating airgap flux density plots


"""
import numpy as np
import matplotlib.pyplot as plt


def airgap(airgap, ax=0):
    """creates plot of flux density in airgap"""
    if ax == 0:
        ax = plt.gca()
    ax.set_title('Airgap Flux Density [T]')
    ax.plot(airgap['pos'], airgap['B'],
            label='Max {:4.2f} T'.format(max(np.abs(airgap['B']))))
    ax.plot(airgap['pos'], airgap['B_fft'],
            label='Base Ampl {:4.2f} T'.format(airgap['Bamp']))
    ax.set_xlabel('Position/°')
    ax.legend()
    ax.grid(True)


def airgap_fft(airgap, bmin=1e-2, ax=0):
    """plot airgap harmonics"""
    unit = 'T'
    if ax == 0:
        ax = plt.gca()
    ax.set_title('Airgap Flux Density Harmonics / {}'.format(unit))
    ax.grid(True)
    order, fluxdens = np.array([(n, b) for n, b in zip(airgap['nue'],
                                                       airgap['B_nue']) if b > bmin]).T
    try:
        markerline1, stemlines1, _ = ax.stem(order, fluxdens, '-.', basefmt=" ",
                                             use_line_collection=True)
        ax.set_xticks(order)
    except ValueError:  # empty sequence
        pass



#!/usr/bin/env python3
#
import unittest
import os
import femagtools.mcv
import numpy as np


class McvWriterTest(unittest.TestCase):
    def test_read_write_mcv(self):
        testPath = os.path.split(__file__)[0]
        if not testPath:
            testPath = '.'
        filename = "data/TKS_NO_20.MCV"
        filename_out = "data/TKS_NO_20_out.MCV"
        reader = femagtools.mcv.Reader()
        reader.readMcv('{0}/{1}'.format(testPath, filename))
        r = reader.get_results()
        self.assertEqual(r['desc'],
                         u'PowerCore\xae NO 20 ;ThyssenKrupp Steel Eur')
        self.assertEqual(len(r['curve'][0]['bi']), 24)
        self.assertEqual(r['curve'][0]['bi'][0], 0.0)
        self.assertAlmostEqual(r['curve'][0]['bi'][-1], 1.836, 3)

        # test mcv writer
        writer = femagtools.mcv.Writer(r)
        # writer.setData(r)
        writeMcvFile = '{0}/{1}'.format(testPath, filename_out)
        writer.writeMcv(writeMcvFile)
        self.assertNotEqual(writer, None)

        # TEST
        reader2 = femagtools.mcv.Reader()
        reader2.readMcv(writeMcvFile)

        for attr in ['version_mc_curve', 'mc1_curves', 'mc1_title']:
            self.assertAlmostEqual(getattr(reader, attr),
                                   getattr(reader2, attr))
        for attr in ['mc1_remz', 'mc1_bsat', 'mc1_bref', 'mc1_fillfac',
                     'fo', 'Bo', 'ch', 'ch_freq', 'cw',
                     'cw_freq', 'b_coeff', 'rho', 'fe_sat_mag']:
            self.assertAlmostEqual(getattr(reader, attr),
                                   getattr(reader2, attr), 3)
        for attr in ['hi', 'bi']:
            self.assertAlmostEqual(reader.curve[0][attr],
                                   reader2.curve[0][attr], 3)

    def test_read_write_mcv_with_losses(self):
        testPath = os.path.split(__file__)[0]
        if not testPath:
            testPath = '.'
        filename = "data/TKM270-50A-LOSS.MCV"
        filename_out = "data/TKS_LOSS.MCV"
        reader = femagtools.mcv.Reader()
        reader.readMcv('{0}/{1}'.format(testPath, filename))
        r = reader.get_results()
        self.assertEqual(r['name'],
                         u'TKM270-50A-LOSS')
        self.assertEqual(len(r['curve'][0]['bi']), 35)
        self.assertAlmostEqual(r['curve'][0]['bi'][-1], 2.638, 3)
        self.assertEqual(set(r['losses'].keys()),
                         set({'B', 'f', 'pfe', 'Bo', 'fo',
                              'cw', 'cw_freq', 'b_coeff', 'ch', 'ch_freq'}))
        self.assertEqual(len(r['losses']['pfe']), 20)
        self.assertEqual(len(r['losses']['pfe'][8]), 19)
        self.assertAlmostEqual(r['losses']['pfe'][8][18], 3097.6, 1)

        # test mcv writer
        writer = femagtools.mcv.Writer(r)
        # writer.setData(r)
        writeMcvFile = '{0}/{1}'.format(testPath, filename_out)
        writer.writeMcv(writeMcvFile)
        self.assertNotEqual(writer, None)

        # TEST
        reader2 = femagtools.mcv.Reader()
        reader2.readMcv(writeMcvFile)

        for attr in ['version_mc_curve', 'mc1_curves', 'mc1_title']:
            self.assertAlmostEqual(getattr(reader, attr),
                                   getattr(reader2, attr))
        for attr in ['mc1_remz', 'mc1_bsat', 'mc1_bref', 'mc1_fillfac',
                     'fo', 'Bo', 'ch', 'ch_freq', 'cw',
                     'cw_freq', 'b_coeff', 'rho', 'fe_sat_mag']:
            self.assertAlmostEqual(getattr(reader, attr),
                                   getattr(reader2, attr), 3)
        for attr in ['hi', 'bi']:
            self.assertAlmostEqual(reader.curve[0][attr],
                                   reader2.curve[0][attr], 3)
        np.testing.assert_almost_equal(reader.losses['f'],
                                       reader2.losses['f'], 5)
        np.testing.assert_almost_equal(reader.losses['B'],
                                       reader2.losses['B'], 5)

if __name__ == '__main__':
    unittest.main()



"""
  Copyright (c) 2015 Semafor Informatik & Energie AG

"""
import sys
import numpy as np
import copy
import operator as op


class Individual:
    def __init__(self, n_dim, f_dim):
        self.cur_f = [0]*f_dim
        self.cur_x = [0]*n_dim
        self.cur_v = []
        self.cur_c = []

        self.rank = 0
        self.crowd_d = 0
        self.idx = 0

    def __str__(self):
        return "f {} x {} rank {} crowd_d {}".format(self.cur_f, self.cur_x,
                                                     self.rank, self.crowd_d)


class Population:

    def __init__(self, probl, size, seed=None):
        self.individuals = []
        self.problem = probl
        self.champion = None
        self.dom_count = []
        self.dom_list = []
        np.random.seed(seed)
        for s in range(size):
            self.append([np.random.uniform(lb, ub)
                         for lb, ub in zip(self.problem.lower,
                                           self.problem.upper)])
#        kmax=int(np.sqrt(size))+1
#        f=[]
#        for i in range(kmax):
#            for k in range(kmax):
#                f.append((i*0.8/kmax +0.1, k*0.8/kmax +0.1))
#        for k,i in enumerate(self.individuals):
#            i.cur_f = f[k]

    def size(self):
        return len(self.individuals)

    def copy(self):
        return copy.copy(self)

    def append(self, child):
        self.individuals.append(Individual(self.problem.dimension,
                                           self.problem.f_dim))
        self.individuals[-1].cur_x = child
        self.individuals[-1].idx = len(self.individuals)-1
        self.init_velocity()

    def eval(self):
        for k, i in enumerate(self.individuals):
            i.cur_f = self.problem.objfun(i.cur_x)
        self.update()

    def best_idx(self):
        return [i.idx for i in sorted(self.individuals,
                                      key=op.attrgetter('rank',
                                                        'crowd_d'))]

    def merge(self, pop):
        "sort by rank and crowding distance (proximity)"
        self.individuals += pop.individuals
        self.update()
        for i in self.individuals:
            if abs(i.crowd_d) > sys.float_info.epsilon:
                i.crowd_d = 1/i.crowd_d
            else:
                i.crowd_d = sys.float_info.max

        best = sorted(self.individuals,
                      key=op.attrgetter('rank',
                                        'crowd_d'))
        self.individuals = best[:pop.size()]
        self.update()

    def init_velocity(self):
        for i, j in enumerate(self.individuals[-1].cur_x):
            w = (self.problem.upper[i] - self.problem.lower[i])/2
            self.individuals[-1].cur_v.append(np.random.uniform(-w, w))

    def populate(self, x, f, s):
        """populate with decision and objective values multiplied by sign
        x: vector of decision values
        f: matrix of objective values
        s: vector of signs (1, -1)"""
        for k, i in enumerate(self.individuals):
            i.cur_x = x[k]
            i.cur_f = [v*sign
                       for v, sign in zip(f[:, k], s)]
        self.update()

    def get_ranked_decisions(self):
        px = dict()
        for i in self.individuals:
            k = i.rank
            if k in px:
                px[k].append(i.cur_x)
            else:
                px[k] = [i.cur_x]
        return px

    def get_ranked_objectives(self, s):
        po = dict()
        for i in self.individuals:
            cur_f = [v*sign
                     for v, sign in zip(i.cur_f, s)]
            k = i.rank
            if k in po:
                po[k].append(cur_f)
            else:
                po[k] = [cur_f]
        return po

    def update(self):
        size = len(self.individuals)
        self.dom_count = []
        self.dom_list = [[] for s in range(size)]
        self.champion = None
        for s in range(size):
            self.dom_count.append(0)
            self.update_dom(s)
            self.update_champion(s)
        self.update_pareto_information()

    def update_dom(self, n):
        """Loop over the population (j) and construct
        dom_list[n] and dom_count """

        for i, x in enumerate(self.individuals):
            if i != n:
                # check if individual in position i
                # dominates the one in position n
                if self.problem.compare_fc(x.cur_f, x.cur_c,
                                           self.individuals[n].cur_f,
                                           self.individuals[n].cur_c):
                    self.dom_count[n] += 1
                    self.dom_list[i].append(n)

    def update_champion(self, idx):
        if self.champion is None or \
            self.problem.compare_fc(self.individuals[idx].cur_f,
                                    self.individuals[idx].cur_c,
                                    self.champion['f'], self.champion['c']):
            self.champion = dict(x=self.individuals[idx].cur_x,
                                 f=self.individuals[idx].cur_f,
                                 c=self.individuals[idx].cur_c)

    def update_crowding(self, F):
        # sort along the fitness dimension
        for i in range(self.problem.f_dim):
            I = sorted(F, key=lambda k: self.individuals[k].cur_f[i],
                       reverse=True)
            self.individuals[I[0]].crowd_d = sys.float_info.max
            self.individuals[I[-1]].crowd_d = sys.float_info.max
            df = (self.individuals[I[-1]].cur_f[i] -
                  self.individuals[I[0]].cur_f[i])
            for j in range(1, len(F)-1):
                if abs(df) > sys.float_info.epsilon:
                    self.individuals[I[j]].crowd_d += (
                        self.individuals[I[j+1]].cur_f[i] -
                        self.individuals[I[j-1]].cur_f[i])/df

    def update_pareto_information(self):
        size = len(self.individuals)
        self.pareto_rank = [0]*size
        F = [i for i, c in enumerate(self.dom_count) if c == 0]
        irank = 1
        dom_count_copy = list(self.dom_count)
        while True:
            self.update_crowding(F)
            S = []
            for i, f in enumerate(F):
                for j, k in enumerate(self.dom_list[f]):
                    dom_count_copy[k] -= 1
                    if dom_count_copy[k] == 0:
                        S.append(k)
                        self.pareto_rank[k] = irank
                        self.individuals[k].rank = irank
            if S:
                F = list(S)
            else:
                return

            irank += 1

    def compute_pareto_fronts(self):
        self.update_pareto_information()
        retval = [[] for s in range(max(self.pareto_rank)+1)]
        for i, j in enumerate(self.individuals):
            retval[self.pareto_rank[i]].append(i)
        return retval

    def compute_ideal(self):
        return [min(x)
                for x in zip(*[i.cur_f
                               for i in self.individuals if i.rank == 0])]

    def compute_nadir(self):
        return [max(x)
                for x in zip(*[i.cur_f
                               for i in self.individuals if i.rank == 0])]

    def compute_worst(self):
        return [max(x)
                for x in zip(*[i.cur_f
                               for i in self.individuals])]

    def compute_norm_dist(self):
        zi = np.array(self.compute_ideal())
        zw = np.array(self.compute_worst())
        znad = np.array(self.compute_nadir())
        return np.sqrt(((znad-zi)**2).sum() / ((zw-zi)**2).sum())

    def plot_pareto_fronts(
            self,
            rgb=(
                0,
                0,
                0),
            comp=[
                0,
                1],
            symbol='o',
            size=6,
            fronts=[]):
        """
        Plots the population pareto front in a 2-D graph

        USAGE: pop.plot_pareto_front(comp = [0,1], rgb=(0,1,0))

        * comp: components of the fitness function to plot in the 2-D window
        * rgb: specify the color of the 1st front (use strong colors here)
        * symbol: marker for the individual
        * size: size of the markersymbol
        * fronts: list of fronts to be plotted (use [0] to only show the first)
        """
        from numpy import linspace
        import matplotlib.pyplot as plt

        if len(comp) != 2:
            raise ValueError(
                'Invalid components of the objective function selected for plot')

        p_dim = self.problem.f_dimension

        if p_dim == 1:
            raise ValueError(
                'Pareto fronts of a 1-dimensional problem cannot be plotted')

        if not all([c in range(0, p_dim) for c in comp]):
            raise ValueError(
                'You need to select valid components of the objective function')

        p_list = self.compute_pareto_fronts()
        if (len(fronts) > 0):
            n = len(p_list)
            consistent = [d < n for d in fronts]
            if consistent.count(False) > 0:
                raise ValueError(
                    'Check your fronts list, there seem to be not enough fronts')
            p_list = [p_list[idx] for idx in fronts]

        cl = list(zip(linspace(0.9 if rgb[0] else 0.1, 0.9, len(p_list)),
                      linspace(0.9 if rgb[1] else 0.1, 0.9, len(p_list)),
                      linspace(0.9 if rgb[2] else 0.1, 0.9, len(p_list))))

        for id_f, f in enumerate(p_list):
            for ind in f:
                plt.plot([ind.cur_f[comp[0]]],
                         [ind.cur_f[comp[1]]],
                         symbol,
                         color=cl[id_f], markersize=size)
            x = [ind.cur_f[comp[0]] for ind in f]
            y = [ind.cur_f[comp[1]] for ind in f]
            tmp = [(a, b) for a, b in zip(x, y)]
            tmp = sorted(tmp, key=lambda k: k[0])
            plt.step([c[0] for c in tmp], [c[1]
                     for c in tmp], color=cl[id_f], where='post')
        return plt.gca()



# -*- coding: utf-8 -*-
"""
    femagtools.parstudy
    ~~~~~~~~~~~~~~~~~~~

    Parameter study simulation



"""
import scipy as sc
import logging
import glob
import os
import time
import shutil
import functools
import pathlib
import numpy as np
import femagtools
import femagtools.model
import femagtools.fsl
import femagtools.condor
import femagtools.moproblem
import femagtools.getset
from .femag import set_magnet_properties

logger = logging.getLogger(__name__)


def get_report(decision_vars, objective_vars, objectives, x):
    """returns a combined list of objective and design values
    with header"""
    y = np.reshape(np.asarray(objectives),
                   (np.shape(objectives)[0], np.shape(x)[0])).T
    c = np.arange(x.shape[0]).reshape(x.shape[0], 1)
    return [[d['label'] for d in decision_vars] +
            [o['label']
             for o in objective_vars] + ['Directory'],
            [d['name'] for d in decision_vars] +
            [o['name']
             for o in objective_vars]] + [
        z[:-1].tolist() + [int(z[-1])]
        for z in np.hstack((x, y, c))]


def baskets(items, basketsize=10):
    """generates balanced baskets from iterable, contiguous items"""
    num_items = len(items)
    num_baskets = max(1, num_items//basketsize)
    if num_items % basketsize and basketsize < num_items:
        num_baskets += 1
    step = num_items//num_baskets
    for i in range(0, num_items, step):
        yield items[i:i+step]


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i+n]


class ParameterStudy(object):
    """abstract base class for parameter variation calculation

    Args:
    workdir: pathname of work directory

    CAUTION: choose the workdir exclusively for this usage.
             do not share it with anything else.
    """

    def __init__(self, workdir,
                 magnetizingCurves=None, magnets=None, condMat=[], result_func=None,
                 repname='grid', cmd=None,
                 templatedirs=[]):  # tasktype='Task'):
        #self.tasktype = tasktype
        self.templatedirs = templatedirs
        self.result_func = result_func
        self.femag = femagtools.Femag(workdir=workdir, cmd=cmd,
                                      magnetizingCurves=magnetizingCurves,
                                      magnets=magnets, condMat=condMat)
        # rudimentary: gives the ability to stop a running parameter variation. thomas.maier/OSWALD
        self.stop = False
        self.reportdir = ''
        self.repname = repname  # prefix for report filename ..-report.csv
        """
        the "owner" of the Grid have to take care to terminate all running xfemag64 or wfemagw64
        processes after setting stop to True
        For example:

        def killFemagThreads():
            if sys.platform.startswith('linux'):
                os.system(
                    "kill $(ps aux | grep '[x]femag64' | awk '{print $2}')")
            else:
                os.system("taskkill /f /im wfemagw64.exe")

        thomas.maier/OSWALD
        """

    def set_report_directory(self, dirname):
        """saves the result files (BCH/BATCH) of every calculation
        into this directory. Throws ValueError if directory is not empty or
        FileNotFoundError if it does not exist.
        Args:
          dirname: name of report directory
        """
        if os.listdir(dirname):
            raise ValueError("directory {} is not empty".format(dirname))
        self.reportdir = dirname

    def setup_model(self, builder, model, recsin=''):
        """builds model in current workdir and returns its filenames"""
        # get and write mag curves
        mc_files = self.femag.copy_magnetizing_curves(model, recsin=recsin)

        if model.is_complete():
            logger.info("setup model in %s", self.femag.workdir)
            filename = 'femag.fsl'
            if 'wdgdef' in model.windings:
                model.windings['wdgfile'] = self.femag.create_wdg_def(
                    model)

            with open(os.path.join(self.femag.workdir, filename), 'w') as f:
                f.write('\n'.join(builder.create_model(
                    model, self.femag.magnets, self.femag.condMat) +
                    ['save_model("close")']))

            self.femag.run(filename)

        model_files = [os.path.join(self.femag.workdir, m)
                       for m in mc_files] + [
            f for sublist in [
                glob.glob(os.path.join(
                    self.femag.workdir,
                    model.name+e))
                for e in (
                    '_*.poc', '*.WID',
                    '.nc', '.[IA]*7')]
            for f in sublist]

        return model_files

    def __call__(self, opt, machine, simulation,
                 engine, bchMapper=None,
                 extra_files=[], num_samples=0):
        """calculate objective vars for all decision vars
        Args:
          opt: variation parameter dict (decision_vars, objective_vars)
          machine: parameter dict of machine
          simulation: parameter dict of simulation
          engine: calculation runner (MultiProc, Condor ..)
          bchMapper: bch result transformation function
          extra_files: list of additional input file names to be copied
          num_samples: number of samples (ingored with Grid sampling)
        """

        self.stop = False  # make sure the calculation will start. thomas.maier/OSWALD

        decision_vars = opt['decision_vars']
        objective_vars = opt.get('objective_vars', {})

        model = femagtools.model.MachineModel(machine)
        builder = femagtools.fsl.Builder(self.templatedirs)

        dvarnames, domain, par_range = self._get_names_and_range(
            decision_vars, num_samples)

        # check if this model needs to be modified
        immutable_model = len([d for d in dvarnames
                               if hasattr(model,
                                          d.split('.')[0])]) == 0

        if isinstance(self.femag.workdir, pathlib.Path):
            workdir = self.femag.workdir
        else:
            workdir = pathlib.Path(self.femag.workdir)
        for d in workdir.glob('[0-9]*'):
            shutil.rmtree(d)

        extra_result_files = []
        if simulation.get('airgap_induc', False):
            extra_result_files.append('bag.dat')
        simulation['arm_length'] = model.lfe
        simulation['lfe'] = model.lfe
        simulation['move_action'] = model.move_action
        simulation['phi_start'] = 0.0
        simulation['range_phi'] = 720/model.get('poles')
        simulation.update(model.windings)
        if 'pocfilename' not in simulation:
            simulation['pocfilename'] = f"{model.name}_{model.poles}p.poc"
        fea = femagtools.model.FeaModel(simulation)

        prob = femagtools.moproblem.FemagMoProblem(decision_vars,
                                                   objective_vars)

        if immutable_model:
            modelfiles = self.setup_model(builder, model, recsin=fea.recsin)
            logger.info("Files %s", modelfiles+extra_files)

        job = engine.create_job(self.femag.workdir)
        # for progress logger
        job.num_cur_steps = fea.get_num_cur_steps()

        f = []
        p = 1
        calcid = 0
        logger.debug(par_range)

        if hasattr(fea, 'poc'):
            fea.poc.pole_pitch = 2*360/model.get('poles')
            fea.pocfilename = fea.poc.filename()
        elapsedTime = 0
        self.bchmapper_data = []  # clear bch data
        # split x value (par_range) array in handy chunks:
        popsize = 0
        for population in baskets(par_range, opt.get('population_size',
                                                     len(par_range))):
            if self.stop:  # try to return the results so far. thomas.maier/OSWALD
                logger.info(
                    'stopping grid execution... returning results so far...')
                try:
                    shape = [len(objective_vars)] + [len(d)
                                                     for d in reversed(domain)]
                    logger.debug("BEFORE: f shape %s --> %s",
                                 np.shape(np.array(f).T), shape)
                    completed = int(functools.reduce(
                        (lambda x, y: x * y), [len(z) for z in domain]))
                    logger.debug("need {} in total".format(completed))
                    remaining = completed - int(np.shape(np.array(f).T)[1])
                    values = int(np.shape(np.array(f).T)[0])
                    logger.debug("going to append {} None values".format(
                        remaining))
                    f += remaining * [values * [np.nan]]
                    shape = [len(objective_vars)] + [len(d)
                                                     for d in reversed(domain)]
                    logger.debug("AFTER: f shape %s --> %s",
                                 np.shape(np.array(f).T), shape)
                    objectives = np.reshape(np.array(f).T, shape)
                    r = dict(f=objectives.tolist(),
                             x=domain)
                    return r
                except:
                    return {}
                    pass
            popsize = max(len(population), popsize)
            logger.info('........ %d / %d results: %s',
                        p, int(np.ceil(len(par_range)/popsize)),
                        np.shape(f))
            job.cleanup()
            for k, x in enumerate(population):
                task = job.add_task(self.result_func)
                for fn in extra_files:
                    task.add_file(fn)
                if immutable_model:
                    prob.prepare(x, [fea, self.femag.magnets])
                    for m in modelfiles:
                        task.add_file(m)
                    set_magnet_properties(model, fea, self.femag.magnets)
                    task.add_file(
                        'femag.fsl',
                        builder.create_open(model) +
                        builder.create_fe_losses(model) +
                        builder.create_analysis(fea) +
                        ['save_model("close")'])

                else:
                    prob.prepare(x, [model, fea, self.femag.magnets])
                    logger.info("prepare %s", x)
                    for mc in self.femag.copy_magnetizing_curves(
                            model,
                            dir=task.directory,
                            recsin=fea.recsin):
                        task.add_file(mc)
                    set_magnet_properties(model, fea, self.femag.magnets)
                    task.add_file(
                        'femag.fsl',
                        builder.create_model(model, self.femag.magnets) +
                        builder.create_analysis(fea) +
                        ['save_model("close")'])

                if hasattr(fea, 'poc'):
                    task.add_file(fea.pocfilename,
                                  fea.poc.content())
                if hasattr(fea, 'stateofproblem'):
                    task.set_stateofproblem(fea.stateofproblem)

            tstart = time.time()
            status = engine.submit(extra_result_files)
            logger.info('Started %s', status)
            status = engine.join()
            tend = time.time()
            elapsedTime += (tend-tstart)
            logger.info("Elapsed time %d s Status %s",
                        (tend-tstart), status)
            for t in job.tasks:
                if t.status == 'C':
                    r = t.get_results()
                    # save result file if requested:
                    if self.reportdir:
                        repdir = os.path.join(self.reportdir,
                                              str(calcid))
                        os.makedirs(repdir)
                        try:
                            shutil.copy(glob.glob(os.path.join(
                                t.directory, r.filename)+'.B*CH')[0], repdir)
                        except (FileNotFoundError, AttributeError):
                            # must be a failure, copy all files
                            for ff in glob.glob(
                                    os.path.join(t.directory, '*')):
                                shutil.copy(ff, repdir)
                        calcid += 1
                    if isinstance(r, dict) and 'error' in r:
                        logger.warn("job %d failed: %s", k, r['error'])
                        if objective_vars:
                            f.append([float('nan')]*len(objective_vars))
                        else:
                            f.append(dict())
                    else:
                        if bchMapper:
                            bchData = bchMapper(r)
                            self.addBchMapperData(bchData)
                            prob.setResult(bchData)
                        elif isinstance(r, dict):
                            prob.setResult(femagtools.getset.GetterSetter(r))
                        else:
                            prob.setResult(r)
                        f.append(prob.objfun([]))
                else:
                    if objective_vars:
                        f.append([float('nan')]*len(objective_vars))
                    else:
                        f.append(dict())
            p += 1

        logger.info('Total elapsed time %d s ...... DONE', elapsedTime)

        try:
            if objective_vars:
                # Note results f must be transposed but may have different shapes
                objectives = list(zip(*f))
            else:
                objectives = f
            if self.reportdir:
                self._write_report(decision_vars, objective_vars,
                                   objectives, par_range)
            return dict(f=objectives,
                        x=domain)
        except ValueError as v:
            logger.error(v)
            return dict(f=f, x=domain)

    def addBchMapperData(self, bchData):
        self.bchmapper_data.append(bchData)

    def getBchMapperData(self):
        return self.bchmapper_data

    def _write_report(self, decision_vars, objective_vars, objectives, par_range):
        with open(os.path.join(self.reportdir, f'{self.repname}-report.csv'), 'w') as f:
            for line in get_report(decision_vars, objective_vars,
                                   objectives, par_range):
                f.write(';'.join([str(v) for v in line]))
                f.write('\n')


class List(ParameterStudy):
    """List Parameter variation calculation"""

    def __init__(self, workdir,
                 magnetizingCurves=None, magnets=None, condMat=[],
                 result_func=None, cmd=None):  # tasktype='Task'):
        super(self.__class__, self).__init__(
            workdir,
            magnetizingCurves, magnets,
            condMat=condMat, result_func=result_func,
            repname='list', cmd=cmd)

    def _get_names_and_range(self, dvars, num_samples):
        if 'list' in dvars:
            par_range = dvars['list']
            domain = [r for r in zip(*par_range)]
            dnames = dvars['columns']
        else:
            domain = [d['values'] for d in dvars]
            par_range = [r for r in zip(*domain)]
            dnames = [d['name'] for d in dvars]
        return dnames, domain, par_range


class LatinHypercube(ParameterStudy):
    """Latin Hypercube sampling parameter variation calculation"""

    def __init__(self, workdir,
                 magnetizingCurves=None, magnets=None, condMat=[],
                 result_func=None, cmd=None):  # tasktype='Task'):
        super(self.__class__, self).__init__(
            workdir,
            magnetizingCurves, magnets,
            condMat=[], result_func=result_func, repname='lhs', cmd=cmd)

    def _get_names_and_range(self, dvars, num_samples):
        dvarnames = [d['name'] for d in dvars]
        l_bounds = [d['bounds'][0] for d in dvars]
        u_bounds = [d['bounds'][1] for d in dvars]

        N = num_samples
        sampler = sc.stats.qmc.LatinHypercube(d=len(l_bounds), scramble=False)
        sample = sampler.random(n=N)
        par_range = sc.stats.qmc.scale(sample, l_bounds, u_bounds)
        domain = par_range.T.tolist()
        return dvarnames, domain, par_range


class Sobol(ParameterStudy):
    """Sobol sampling parameter variation calculation"""

    def __init__(self, workdir,
                 magnetizingCurves=None, magnets=None, condMat=[],
                 result_func=None, cmd=None):  # tasktype='Task'):
        super(self.__class__, self).__init__(workdir,
                                             magnetizingCurves, magnets,
                                             condMat=condMat, result_func=result_func,
                                             repname='sobol', cmd=cmd)

    def _get_names_and_range(self, dvars, num_samples):
        dvarnames = [d['name'] for d in dvars]
        l_bounds = [d['bounds'][0] for d in dvars]
        u_bounds = [d['bounds'][1] for d in dvars]

        N = num_samples
        sampler = sc.stats.qmc.Sobol(d=len(l_bounds), scramble=False)
        sample = sampler.random_base2(m=round(np.log(N)/np.log(2)))
        par_range = sc.stats.qmc.scale(sample, l_bounds, u_bounds)
        domain = par_range.T.tolist()
        return dvarnames, domain, par_range


class Grid(ParameterStudy):
    """Grid Parameter variation calculation"""

    def __init__(self, workdir,
                 magnetizingCurves=None, magnets=None, condMat=[],
                 result_func=None, cmd=None):  # tasktype='Task'):
        super(self.__class__, self).__init__(
            workdir,
            magnetizingCurves, magnets, condMat=condMat,
            result_func=result_func, cmd=cmd)

    def __create_parameter_range(self, domain):
        """returns the transposed array of the combined domain values"""
        L = [len(d) for d in domain]
        LS = np.prod(L)
        s = []
        e = 1
        for d in domain:
            LS = LS//len(d)
            s.append(np.repeat(d*LS, e))
            e = e*L[0]
            L = L[1:]
        return np.array(s).T

    def _get_names_and_range(self, dvars, num_samples):
        dvarnames = [d['name'] for d in dvars]
        domain = [list(np.linspace(d['bounds'][0],
                                   d['bounds'][1],
                                   d['steps']))
                  for d in dvars]

        par_range = self.__create_parameter_range(domain)
        return dvarnames, [r for r in zip(*par_range)], par_range



"""
    femagtools.plot.wdg
    ~~~~~~~~~~~~~~~~~~~

    Creating winding plots

"""
import numpy as np
import matplotlib.pyplot as plt


def mmf(f, title='', ax=0):
    """plot magnetomotive force (mmf) of winding"""
    if ax == 0:
        ax = plt.gca()
    if title:
        ax.set_title(title)
    ax.plot(np.array(f['pos'])/np.pi*180, f['mmf'])
    ax.plot(np.array(f['pos_fft'])/np.pi*180, f['mmf_fft'])
    ax.set_xlabel('Position / Deg')

    phi = [f['alfa0']/np.pi*180, f['alfa0']/np.pi*180]
    y = [min(f['mmf_fft']), 1.1*max(f['mmf_fft'])]
    ax.plot(phi, y, '--')
    alfa0 = round(f['alfa0']/np.pi*180, 2)
    ax.text(phi[0]/2, y[0]+0.05, f"{alfa0}°",
            ha="center", va="bottom")
    ax.annotate(f"", xy=(phi[0], y[0]),
                xytext=(0, y[0]), arrowprops=dict(arrowstyle="->"))
    ax.grid()


def mmf_fft(f, title='', mmfmin=1e-2, ax=0):
    """plot winding mmf harmonics"""
    if ax == 0:
        ax = plt.gca()
    if title:
        ax.set_title(title)
    else:
        ax.set_title('MMF Harmonics (per phase)')
    ax.grid(True)
    order, mmf = np.array([(n, m) for n, m in zip(f['nue'],
                                                  f['mmf_nue']) if m > mmfmin]).T
    try:
        markerline1, stemlines1, _ = ax.stem(order, mmf, '-.', basefmt=" ",
                                             use_line_collection=True)
        ax.set_xticks(order)
    except ValueError:  # empty sequence
        pass


def zoneplan(wdg, ax=0):
    """plot zone plan of winding wdg"""
    from matplotlib.patches import Rectangle
    upper, lower = wdg.zoneplan()
    Qb = len([n for l in upper for n in l])
    from femagtools.windings import coil_color
    rh = 0.5
    if lower:
        yl = rh
        ymax = 2*rh + 0.2
    else:
        yl = 0
        ymax = rh + 0.2
    if ax == 0:
        ax = plt.gca()
    ax.axis('off')
    ax.set_xlim([-0.5, Qb-0.5])
    ax.set_ylim([0, ymax])
    ax.set_aspect(Qb/6+0.3)

    for i, p in enumerate(upper):
        for x in p:
            ax.add_patch(Rectangle((abs(x)-1.5, yl), 1, rh,
                                   facecolor=coil_color[i],
                                   edgecolor='white', fill=True))
            s = f'+{i+1}' if x > 0 else f'-{i+1}'
            ax.text(abs(x)-1, yl+rh/2, s, color='black',
                    ha="center", va="center")
    for i, p in enumerate(lower):
        for x in p:
            ax.add_patch(Rectangle((abs(x)-1.5, yl-rh), 1, rh,
                                   facecolor=coil_color[i],
                                   edgecolor='white', fill=True))
            s = f'+{i+1}' if x > 0 else f'-{i+1}'
            ax.text(abs(x)-1, yl-rh/2, s, color='black',
                    ha="center", va="center")

    yu = yl+rh
    step = 1 if Qb < 25 else 2
    if lower:
        yl -= rh
    margin = 0.05
    ax.text(-0.5, yu+margin, f'Q={wdg.Q}, p={wdg.p}, q={round(wdg.q,4)}',
            ha='left', va='bottom', size=15)
    for i in range(0, Qb, step):
        ax.text(i, yl-margin, f'{i+1}', ha="center", va="top")


def winding_factors(wdg, n=8, ax=0):
    """plot winding factors"""
    ax = plt.gca()
    ax.set_title(f'Winding factors Q={wdg.Q}, p={wdg.p}, q={round(wdg.q,4)}')
    ax.grid(True)
    order, kwp, kwd, kw = np.array(
        [(n, k1, k2, k3)
         for n, k1, k2, k3 in zip(wdg.kw_order(n),
                                  wdg.kwp(n),
                                  wdg.kwd(n),
                                  wdg.kw(n))]).T
    try:
        markerline1, stemlines1, _ = ax.stem(order-1, kwp,
                                             'C1:', basefmt=" ",
                                             markerfmt='C1.',
                                             use_line_collection=True,
                                             label='Pitch')
        markerline2, stemlines2, _ = ax.stem(order+1, kwd,
                                             'C2:', basefmt=" ",
                                             markerfmt='C2.',
                                             use_line_collection=True,
                                             label='Distribution')
        markerline3, stemlines3, _ = ax.stem(order, kw,
                                             'C0-', basefmt=" ",
                                             markerfmt='C0o',
                                             use_line_collection=True,
                                             label='Total')
        ax.set_xticks(order)
        ax.legend()
    except ValueError:  # empty sequence
        pass


def winding(wdg, ax=0):
    """plot coils of windings wdg"""
    from matplotlib.patches import Rectangle
    from matplotlib.lines import Line2D
    from femagtools.windings import coil_color

    coil_len = 25
    coil_height = 4
    dslot = 8
    arrow_head_length = 2
    arrow_head_width = 2

    if ax == 0:
        ax = plt.gca()
    z = wdg.zoneplan()
    xoff = 0
    if z[-1]:
        xoff = 0.75
    yd = dslot*wdg.yd
    mh = 2*coil_height/yd
    slots = sorted([abs(n) for m in z[0] for n in m])
    smax = slots[-1]*dslot
    for n in slots:
        x = n*dslot
        ax.add_patch(Rectangle((x + dslot/4, 1), dslot /
                     2, coil_len - 2, fc="lightblue"))
        ax.text(x, coil_len / 2,
                str(n),
                horizontalalignment="center",
                verticalalignment="center",
                backgroundcolor="white",
                bbox=dict(boxstyle='circle,pad=0', fc="white", lw=0))
    line_thickness = [0.6, 1.2]
    for i, layer in enumerate(z):
        b = -xoff if i else xoff
        lw = line_thickness[i]
        direction = ['right', 'left']
        d = 1
        for m, mslots in enumerate(layer):
            for k in mslots:
                x = abs(k) * dslot + b
                xpoints = []
                ypoints = []
                if wdg.q >= 1 or wdg.l > 1:
                    if (i == 0 and (k > 0 or (k < 0 and wdg.l > 1))):
                        d = 0  # right
                    else:
                        d = 1  # left
                elif d == 0:
                    d = 1
                else:
                    d = 0
                if direction[d] == 'right':
                    # first layer, positive dir or neg. dir and 2-layers:
                    #   from right bottom
                    if x + yd > smax+b:
                        dx = dslot if yd > dslot else yd/4
                        xpoints = [x + yd//2 + dx - xoff]
                        ypoints = [-coil_height + mh*dx]
                    xpoints += [x + yd//2 - xoff, x, x, x + yd//2-xoff]
                    ypoints += [-coil_height, 0, coil_len,
                                coil_len+coil_height]
                    if x + yd > smax+b:
                        xpoints += [x + yd//2 + dx - xoff]
                        ypoints += [coil_len+coil_height - mh*dx]
                else:
                    # from left bottom
                    if x - yd < 0:  # and x - yd/2 > -3*dslot:
                        dx = dslot if yd > dslot else yd/4
                        xpoints = [x - yd//2 - dx + xoff]
                        ypoints = [- coil_height + mh*dx]
                    xpoints += [x - yd//2+xoff, x, x, x - yd/2+xoff]
                    ypoints += [-coil_height, 0, coil_len,
                                coil_len+coil_height]
                    if x - yd < 0:  # and x - yd > -3*dslot:
                        xpoints += [x - yd//2 - dx + xoff]
                        ypoints += [coil_len + coil_height - mh*dx]

                ax.add_line(Line2D(xpoints, ypoints,
                            color=coil_color[m], lw=lw))

                if k > 0:
                    h = arrow_head_length
                    y = coil_len * 0.8
                else:
                    h = -arrow_head_length
                    y = coil_len * 0.2
                ax.arrow(x, y, 0, h,
                         length_includes_head=True,
                         head_starts_at_zero=False,
                         head_length=arrow_head_length,
                         head_width=arrow_head_width,
                         fc=coil_color[m], lw=0)
    if False:  # TODO show winding connections
        m = 0
        for k in [n*wdg.Q/wdg.p/wdg.m + 1 for n in range(wdg.m)]:
            if k < len(slots):
                x = k * dslot + b + yd/2 - xoff
                ax.add_line(Line2D([x, x],
                                   [-2*coil_height, -coil_height],
                                   color=coil_color[m], lw=lw))
                ax.text(x, -2*coil_height+0.5, str(m+1), color=coil_color[m])
            m += 1
    ax.autoscale(enable=True)
    ax.set_axis_off()
